@@ -25,6 +25,8 @@ use clap::ValueEnum;
 use sequoia_openpgp as openpgp;
 use openpgp::armor;
 use openpgp::fmt::hex;
+use openpgp::KeyHandle;
+use openpgp::parse::Cookie;
 use openpgp::serialize::stream::Armorer;
 use openpgp::serialize::stream::Message;
 use openpgp::types::KeyFlags;
@@ -104,10 +106,10 @@ pub struct FileOrStdin(Option<PathBuf>);
 
 impl ClapData for FileOrStdin {
     const VALUE_NAME: &'static str = "FILE";
-    const HELP_REQUIRED: &'static str =
-        "Reads from FILE or stdin if FILE is '-'";
     const HELP_OPTIONAL: &'static str =
-        "Reads from FILE or stdin if omitted";
+        "Read from FILE or stdin if FILE is '-'";
+    const HELP_REQUIRED: &'static str =
+        "Read from FILE or stdin if omitted";
 }
 
 impl FileOrStdin {
@@ -132,13 +134,14 @@ impl FileOrStdin {
     /// Get a boxed BufferedReader for the FileOrStdin
     ///
     /// Opens a file if there is Some(PathBuf), else opens stdin.
-    pub fn open(&self) -> Result<Box<dyn BufferedReader<()>>> {
+    pub fn open<'a>(&self) -> Result<Box<dyn BufferedReader<Cookie> + 'a>> {
         if let Some(path) = self.inner() {
             Ok(Box::new(
-                File::open(path)
+                File::with_cookie(path, Default::default())
                 .with_context(|| format!("Failed to open {}", self))?))
         } else {
-            Ok(Box::new(Generic::new(stdin(), None)))
+            Ok(Box::new(
+                Generic::with_cookie(stdin(), None, Default::default())))
         }
     }
 }
@@ -238,13 +241,76 @@ pub struct FileOrCertStore{}
 impl ClapData for FileOrCertStore {
     const VALUE_NAME: &'static str = "FILE";
     const HELP_REQUIRED: &'static str
-        = "Writes to FILE (or stdout if FILE is '-') instead of \
+        = "Write to FILE (or stdout if FILE is '-') instead of \
           importing into the certificate store";
     const HELP_OPTIONAL: &'static str
-        = "Writes to FILE (or stdout when omitted) instead of \
+        = "Write to FILE (or stdout when omitted) instead of \
           importing into the certificate store";
 }
 
+/// Designates a certificate by path, by stdin, or by key handle.
+///
+/// Use [`Sq::lookup_one`] to read the certificate.
+#[derive(Debug)]
+pub enum FileStdinOrKeyHandle {
+    FileOrStdin(FileOrStdin),
+    KeyHandle(KeyHandle),
+}
+
+impl From<FileOrStdin> for FileStdinOrKeyHandle {
+    fn from(file: FileOrStdin) -> Self {
+        FileStdinOrKeyHandle::FileOrStdin(file)
+    }
+}
+
+impl From<&str> for FileStdinOrKeyHandle {
+    fn from(path: &str) -> Self {
+        PathBuf::from(path).into()
+    }
+}
+
+impl From<&Path> for FileStdinOrKeyHandle {
+    fn from(path: &Path) -> Self {
+        path.to_path_buf().into()
+    }
+}
+
+impl From<PathBuf> for FileStdinOrKeyHandle {
+    fn from(path: PathBuf) -> Self {
+        FileStdinOrKeyHandle::FileOrStdin(path.into())
+    }
+}
+
+impl From<&KeyHandle> for FileStdinOrKeyHandle {
+    fn from(kh: &KeyHandle) -> Self {
+        FileStdinOrKeyHandle::KeyHandle(kh.clone())
+    }
+}
+
+impl From<KeyHandle> for FileStdinOrKeyHandle {
+    fn from(kh: KeyHandle) -> Self {
+        FileStdinOrKeyHandle::KeyHandle(kh)
+    }
+}
+
+impl FileStdinOrKeyHandle {
+    /// Returns whether this contains a `FileOrStdin`.
+    pub fn is_file(&self) -> bool {
+        match self {
+            FileStdinOrKeyHandle::FileOrStdin(_) => true,
+            FileStdinOrKeyHandle::KeyHandle(_) => false,
+        }
+    }
+
+    /// Returns whether this contains a `KeyHandle`.
+    pub fn is_key_handle(&self) -> bool {
+        match self {
+            FileStdinOrKeyHandle::FileOrStdin(_) => false,
+            FileStdinOrKeyHandle::KeyHandle(_) => true,
+        }
+    }
+}
+
 /// A type wrapping an optional PathBuf to use as stdout or file output
 ///
 /// When creating `FileOrStdout` from `&str`, providing a `"-"` is interpreted
@@ -281,9 +347,9 @@ pub struct FileOrStdout {
 impl ClapData for FileOrStdout {
     const VALUE_NAME: &'static str = "FILE";
     const HELP_REQUIRED: &'static str =
-        "Writes to FILE or stdout if FILE is '-'";
+        "Write to FILE or stdout if FILE is '-'";
     const HELP_OPTIONAL: &'static str =
-        "Writes to FILE or stdout if omitted";
+        "Write to FILE or stdout if omitted";
 }
 
 impl FileOrStdout {
@@ -586,13 +652,13 @@ impl Default for MetadataTime {
     }
 }
 
-/// Expiry information
+/// Expiration information
 ///
 /// This enum tracks expiry information either in the form of a timestamp or
 /// a duration.
 #[derive(Debug, Clone, Eq, PartialEq)]
 #[non_exhaustive]
-pub enum Expiry {
+pub enum Expiration {
     /// An expiry timestamp
     Timestamp(Time),
     /// A validity duration
@@ -601,28 +667,28 @@ pub enum Expiry {
     Never,
 }
 
-impl Expiry {
-    /// Create a new Expiry in a Result
+impl Expiration {
+    /// Create a new Expiration in a Result
     ///
     /// If `expiry` ends with `"y"`, `"m"`, `"w"`, `"w"`, `"d"` or `"s"` it
     /// is treated as a duration, which is parsed using `parse_duration()` and
-    /// returned in an `Expiry::Duration`.
+    /// returned in an `Expiration::Duration`.
     /// If the special keyword `"never"` is provided as `expiry`,
-    /// `Expiry::Never` is returned.
+    /// `Expiration::Never` is returned.
     /// If `expiry` is an ISO 8601 compatible string it is returned as
-    /// `cli::types::Time` in an `Expiry::Timestamp`.
+    /// `cli::types::Time` in an `Expiration::Timestamp`.
     pub fn new(expiry: &str) -> Result<Self> {
         match expiry {
-            "never" => Ok(Expiry::Never),
+            "never" => Ok(Expiration::Never),
             _ if expiry.ends_with("y")
                 || expiry.ends_with("m")
                 || expiry.ends_with("w")
                 || expiry.ends_with("d")
                 || expiry.ends_with("s") =>
             {
-                Ok(Expiry::Duration(Expiry::parse_duration(expiry)?))
+                Ok(Expiration::Duration(Expiration::parse_duration(expiry)?))
             }
-            _ => Ok(Expiry::Timestamp(Time::from_str(expiry)?)),
+            _ => Ok(Expiration::Timestamp(Time::from_str(expiry)?)),
         }
     }
 
@@ -634,7 +700,7 @@ impl Expiry {
     fn parse_duration(expiry: &str) -> Result<Duration> {
         if expiry.len() < 2 {
             return Err(anyhow::anyhow!(
-                "Expiry must contain at least one digit and one factor."
+                "Expiration must contain at least one digit and one factor."
             ));
         }
 
@@ -651,7 +717,7 @@ impl Expiry {
                     }
                     Ok(count) => count as u64,
                     Err(err) => return Err(err).context(
-                        format!("Expiry '{}' is out of range", digits)
+                        format!("Expiration '{}' is out of range", digits)
                     ),
                 } * match expiry.chars().last() {
                     Some('y') => SECONDS_IN_YEAR,
@@ -660,7 +726,7 @@ impl Expiry {
                     Some('d') => SECONDS_IN_DAY,
                     Some('s') => 1,
                     _ => unreachable!(
-                        "Expiry without 'y', 'm', 'w', 'd' or 's' \
+                        "Expiration without 'y', 'm', 'w', 'd' or 's' \
                                 suffix impossible since checked for it."
                     ),
                 },
@@ -690,42 +756,65 @@ impl Expiry {
     /// Return the expiry as an optional Duration in a Result
     ///
     /// This method returns an Error if the reference time is later than the
-    /// time provided in an `Expiry::Timestamp(Time)`.
+    /// time provided in an `Expiration::Timestamp(Time)`.
     ///
-    /// If self is `Expiry::Timestamp(Time)`, `reference` is used as the start
+    /// If self is `Expiration::Timestamp(Time)`, `reference` is used as the start
     /// of a period, `Some(Time - reference)` is returned.
-    /// If self is `Expiry::Duration(duration)`, `Some(duration)` is returned.
-    /// If self is `Expiry::Never`, `None` is returned.
+    /// If self is `Expiration::Duration(duration)`, `Some(duration)` is returned.
+    /// If self is `Expiration::Never`, `None` is returned.
     pub fn as_duration(
         &self,
         reference: DateTime<Utc>,
     ) -> Result<Option<Duration>> {
         match self {
-            Expiry::Timestamp(time) => Ok(
+            Expiration::Timestamp(time) => Ok(
                 Some(
                     SystemTime::from(time.time).duration_since(reference.into())?
                 )
             ),
-            Expiry::Duration(duration) => Ok(Some(duration.clone())),
-            Expiry::Never => Ok(None),
+            Expiration::Duration(duration) => Ok(Some(duration.clone())),
+            Expiration::Never => Ok(None),
+        }
+    }
+
+    /// Return the expiry as absolute time.
+    pub fn to_systemtime(&self, now: SystemTime) -> Option<SystemTime> {
+        match self {
+            Expiration::Timestamp(t) => Some(t.clone().into()),
+            Expiration::Duration(d) => Some(now + *d),
+            Expiration::Never => None,
         }
     }
 }
 
-impl FromStr for Expiry {
+impl FromStr for Expiration {
     type Err = anyhow::Error;
 
-    fn from_str(s: &str) -> Result<Expiry> {
-        Expiry::new(s)
+    fn from_str(s: &str) -> Result<Expiration> {
+        Expiration::new(s)
     }
 }
 
-impl Display for Expiry {
+impl Display for Expiration {
     fn fmt(&self, f: &mut Formatter) -> std::fmt::Result {
         match self {
-            Expiry::Timestamp(time) => write!(f, "{}", time),
-            Expiry::Duration(duration) => write!(f, "{:?}", duration),
-            Expiry::Never => write!(f, "never"),
+            Expiration::Timestamp(time) => write!(f, "{}", time),
+            Expiration::Duration(duration) => {
+                let seconds = duration.as_secs();
+
+                if seconds % SECONDS_IN_YEAR == 0 {
+                    write!(f, "{}y", seconds / SECONDS_IN_YEAR)
+                } else if seconds % (SECONDS_IN_YEAR / 12) == 0 {
+                    write!(f, "{}m", seconds / (SECONDS_IN_YEAR / 12))
+                } else if seconds % (SECONDS_IN_DAY * 7) == 0 {
+                    write!(f, "{}w", seconds / (SECONDS_IN_DAY * 7))
+                } else if seconds % SECONDS_IN_DAY == 0 {
+                    write!(f, "{}d", seconds / SECONDS_IN_DAY)
+                } else {
+                    write!(f, "{}s", seconds)
+                }
+            },
+            Expiration::Never => write!(f, "never"),
         }
     }
 }
@@ -1025,29 +1114,29 @@ mod test {
     #[test]
     fn test_expiry() {
         assert_eq!(
-            Expiry::new("1y").unwrap(),
-            Expiry::Duration(Duration::new(SECONDS_IN_YEAR, 0)),
+            Expiration::new("1y").unwrap(),
+            Expiration::Duration(Duration::new(SECONDS_IN_YEAR, 0)),
         );
         assert_eq!(
-            Expiry::new("2023-05-15T20:00:00Z").unwrap(),
-            Expiry::Timestamp(Time::from_str("2023-05-15T20:00:00Z").unwrap()),
+            Expiration::new("2023-05-15T20:00:00Z").unwrap(),
+            Expiration::Timestamp(Time::from_str("2023-05-15T20:00:00Z").unwrap()),
         );
         assert_eq!(
-            Expiry::new("never").unwrap(),
-            Expiry::Never,
+            Expiration::new("never").unwrap(),
+            Expiration::Never,
         );
     }
 
     #[test]
     fn test_expiry_parse_duration() {
         assert_eq!(
-            Expiry::parse_duration("1y").unwrap(),
+            Expiration::parse_duration("1y").unwrap(),
             Duration::new(SECONDS_IN_YEAR, 0),
         );
-        assert!(Expiry::parse_duration("f").is_err());
-        assert!(Expiry::parse_duration("-1y").is_err());
-        assert!(Expiry::parse_duration("foo").is_err());
-        assert!(Expiry::parse_duration("1o").is_err());
+        assert!(Expiration::parse_duration("f").is_err());
+        assert!(Expiration::parse_duration("-1y").is_err());
+        assert!(Expiration::parse_duration("foo").is_err());
+        assert!(Expiration::parse_duration("1o").is_err());
     }
 
     #[test]
@@ -1057,7 +1146,7 @@ mod test {
             Utc,
         );
 
-        let expiry = Expiry::Timestamp(
+        let expiry = Expiration::Timestamp(
             Time::try_from(DateTime::from_utc(
                 NaiveDateTime::from_timestamp_opt(2, 0).unwrap(),
                 Utc
@@ -1067,13 +1156,13 @@ mod test {
             Some(Duration::new(1, 0)),
         );
 
-        let expiry = Expiry::Duration(Duration::new(2,0));
+        let expiry = Expiration::Duration(Duration::new(2,0));
         assert_eq!(
             expiry.as_duration(reference).unwrap(),
             Some(Duration::new(2, 0)),
         );
 
-        let expiry = Expiry::Never;
+        let expiry = Expiration::Never;
         assert_eq!(expiry.as_duration(reference).unwrap(), None);
     }
 
@@ -1083,7 +1172,7 @@ mod test {
             NaiveDateTime::from_timestamp_opt(2, 0).unwrap(),
             Utc,
         );
-        let expiry = Expiry::Timestamp(
+        let expiry = Expiration::Timestamp(
             Time::try_from(DateTime::from_utc(
                 NaiveDateTime::from_timestamp_opt(1, 0).unwrap(),
                 Utc
