@@ -1,32 +1,28 @@
-use anyhow::anyhow;
-use std::borrow::Borrow;
 use std::collections::{HashMap, HashSet, btree_map::{BTreeMap, Entry}};
 use std::time::SystemTime;
 
-use sequoia_net::pks;
 use sequoia_openpgp as openpgp;
 use openpgp::cert::prelude::*;
-use openpgp::crypto::{self, Password};
 use openpgp::{Cert, Fingerprint, KeyID, Result};
 use openpgp::packet::prelude::*;
 use openpgp::parse::stream::*;
 use openpgp::policy::HashAlgoSecurity;
 use openpgp::policy::Policy;
-use openpgp::types::KeyFlags;
-use openpgp::types::RevocationStatus;
+use openpgp::types::{
+    AEADAlgorithm,
+    SymmetricAlgorithm,
+};
 
 use sequoia_cert_store as cert_store;
 use cert_store::Store;
 
 use sequoia_wot::store::Store as _;
 
-use crate::{
-    Config,
-    common::password,
-};
+use crate::Sq;
 
 use crate::cli::encrypt::CompressionMode;
 use crate::cli::types::FileOrStdout;
+use crate::cli::{SqCommand, SqSubcommands};
 
 pub mod autocrypt;
 pub mod cert;
@@ -35,218 +31,46 @@ pub mod encrypt;
 pub mod sign;
 pub mod inspect;
 pub mod key;
-pub mod keyring;
-pub mod net;
+pub mod network;
 pub mod pki;
 pub mod toolbox;
 pub mod verify;
 pub mod version;
 
-#[derive(Debug, Clone, PartialEq, Eq)]
-pub enum GetKeysOptions {
-    AllowNotAlive,
-    AllowRevoked,
-}
-
-enum KeyType {
-    Primary,
-    KeyFlags(KeyFlags),
-}
-
-/// Returns suitable signing keys from a given list of Certs.
-fn get_keys<C>(certs: &[C], p: &dyn Policy,
-               private_key_store: Option<&str>,
-               timestamp: Option<SystemTime>,
-               keytype: KeyType,
-               options: Option<&[GetKeysOptions]>)
-    -> Result<Vec<(Box<dyn crypto::Signer + Send + Sync>, Option<Password>)>>
-    where C: Borrow<Cert>
+/// Dispatches the top-level subcommand.
+pub fn dispatch(sq: Sq, command: SqCommand) -> Result<()>
 {
-    let mut bad = Vec::new();
+    match command.subcommand {
+        SqSubcommands::Encrypt(command) =>
+            encrypt::dispatch(sq, command),
+        SqSubcommands::Decrypt(command) =>
+            decrypt::dispatch(sq, command),
+        SqSubcommands::Sign(command) =>
+            sign::dispatch(sq, command),
+        SqSubcommands::Verify(command) =>
+            verify::dispatch(sq, command),
 
-    let options = options.unwrap_or(&[][..]);
-    let allow_not_alive = options.contains(&GetKeysOptions::AllowNotAlive);
-    let allow_revoked = options.contains(&GetKeysOptions::AllowRevoked);
+        SqSubcommands::Inspect(command) =>
+            inspect::dispatch(sq, command),
 
-    let mut keys: Vec<(Box<dyn crypto::Signer + Send + Sync>, Option<Password>)> = vec![];
-    'next_cert: for tsk in certs {
-        let tsk = tsk.borrow();
-        let vc = match tsk.with_policy(p, timestamp) {
-            Ok(vc) => vc,
-            Err(err) => {
-                return Err(
-                    err.context(format!("Found no suitable key on {}", tsk)));
-            }
-        };
+        SqSubcommands::Cert(command) =>
+            cert::dispatch(sq, command),
+        SqSubcommands::Key(command) =>
+            key::dispatch(sq, command),
 
-        let keyiter = match keytype {
-            KeyType::Primary => {
-                Box::new(
-                    std::iter::once(
-                        vc.keys()
-                            .next()
-                            .expect("a valid cert has a primary key")))
-                    as Box<dyn Iterator<Item=ValidErasedKeyAmalgamation<openpgp::packet::key::PublicParts>>>
-            },
-            KeyType::KeyFlags(ref flags) => {
-                Box::new(vc.keys().key_flags(flags.clone()))
-                    as Box<dyn Iterator<Item=_>>
-            },
-        };
-        for ka in keyiter {
-            let bad_ = [
-                ! allow_not_alive && matches!(ka.alive(), Err(_)),
-                ! allow_revoked && matches!(ka.revocation_status(),
-                                            RevocationStatus::Revoked(_)),
-                ! ka.pk_algo().is_supported(),
-            ];
-            if bad_.iter().any(|x| *x) {
-                bad.push((ka.fingerprint(), bad_));
-                continue;
-            }
+        SqSubcommands::Pki(command) =>
+            pki::dispatch(sq, command),
 
-            let key = ka.key();
+        SqSubcommands::Autocrypt(command) =>
+            autocrypt::dispatch(sq, &command),
+        SqSubcommands::Network(command) =>
+            network::dispatch(sq, command),
+        SqSubcommands::Toolbox(command) =>
+            toolbox::dispatch(sq, command),
 
-            if let Some(secret) = key.optional_secret() {
-                let (unencrypted, password) = match secret {
-                    SecretKeyMaterial::Encrypted(ref e) => {
-                        // try passwords from already existing keys
-                        match keys.iter().find_map(|(_, password)| {
-                            password.as_ref().and_then(
-                                |p| e.decrypt(key.pk_algo(), p).ok()
-                                    .map(|key| (key, p.clone())))
-                        }) {
-                            Some((unencrypted, password)) =>
-                                (unencrypted, Some(password)),
-                            None => {
-                                let password = password::prompt_to_unlock(
-                                    &format!("key {}/{}", tsk, key))?;
-                                (
-                                    e.decrypt(key.pk_algo(), &password)
-                                        .map_err(|_| anyhow!("Incorrect password."))?,
-                                    Some(password),
-                                )
-                            }
-                        }
-                    }
-                    SecretKeyMaterial::Unencrypted(ref u) => (u.clone(), None),
-                };
-
-                keys.push((
-                    Box::new(
-                        crypto::KeyPair::new(key.clone(), unencrypted).unwrap()
-                    ),
-                    password,
-                ));
-                continue 'next_cert;
-            } else if let Some(private_key_store) = private_key_store {
-                let input_password = password::prompt_to_unlock(
-                    &format!("key {}/{}", tsk, key))?;
-                match pks::unlock_signer(private_key_store, key.clone(), &input_password) {
-                    Ok(signer) => {
-                        keys.push((signer, Some(input_password.clone())));
-                        continue 'next_cert;
-                    },
-                    Err(error) => wprintln!("Could not unlock key: {:?}", error),
-                }
-            }
-        }
-
-        let timestamp = timestamp.map(|t| {
-            chrono::DateTime::<chrono::offset::Utc>::from(t)
-        });
-
-        let mut context = Vec::new();
-        for (fpr, [not_alive, revoked, not_supported]) in bad {
-            let id: String = if fpr == tsk.fingerprint() {
-                fpr.to_string()
-            } else {
-                format!("{}/{}", tsk.fingerprint(), fpr)
-            };
-
-            let preface = if let Some(t) = timestamp {
-                format!("{} was not considered because\n\
-                         at the specified time ({}) it was",
-                        id, t)
-            } else {
-                format!("{} was not considered because\nit is", fpr)
-            };
-
-            let mut reasons = Vec::new();
-            if not_alive {
-                reasons.push("not alive");
-            }
-            if revoked {
-                reasons.push("revoked");
-            }
-            if not_supported {
-                reasons.push("not supported");
-            }
-
-            context.push(format!("{}: {}",
-                                 preface, reasons.join(", ")));
-        }
-
-        if context.is_empty() {
-            return Err(anyhow::anyhow!(
-                format!("Found no suitable key on {}", tsk)));
-        } else {
-            let context = context.join("\n");
-            return Err(
-                anyhow::anyhow!(
-                    format!("Found no suitable key on {}", tsk))
-                    .context(context));
-        }
+        SqSubcommands::Version(command) =>
+            version::dispatch(sq, command),
     }
-
-    Ok(keys)
-}
-
-/// Returns the primary keys from a given list of Certs.
-///
-/// This returns one key for each Cert.  If a Cert doesn't have an
-/// appropriate key, then this returns an error.
-fn get_primary_keys<C>(certs: &[C], p: &dyn Policy,
-                       private_key_store: Option<&str>,
-                       timestamp: Option<SystemTime>,
-                       options: Option<&[GetKeysOptions]>)
-    -> Result<Vec<(Box<dyn crypto::Signer + Send + Sync>, Option<Password>)>>
-    where C: std::borrow::Borrow<Cert>
-{
-    get_keys(certs, p, private_key_store, timestamp,
-             KeyType::Primary, options)
-}
-
-/// Returns suitable signing keys from a given list of Certs.
-///
-/// This returns one key for each Cert.  If a Cert doesn't have an
-/// appropriate key, then this returns an error.
-fn get_signing_keys<C>(certs: &[C], p: &dyn Policy,
-                       private_key_store: Option<&str>,
-                       timestamp: Option<SystemTime>,
-                       options: Option<&[GetKeysOptions]>)
-    -> Result<Vec<(Box<dyn crypto::Signer + Send + Sync>, Option<Password>)>>
-    where C: Borrow<Cert>
-{
-    get_keys(certs, p, private_key_store, timestamp,
-             KeyType::KeyFlags(KeyFlags::empty().set_signing()),
-             options)
-}
-
-/// Returns suitable certification keys from a given list of Certs.
-///
-/// This returns one key for each Cert.  If a Cert doesn't have an
-/// appropriate key, then this returns an error.
-pub fn get_certification_keys<C>(certs: &[C], p: &dyn Policy,
-                             private_key_store: Option<&str>,
-                             timestamp: Option<SystemTime>,
-                             options: Option<&[GetKeysOptions]>)
-    -> Result<Vec<(Box<dyn crypto::Signer + Send + Sync>, Option<Password>)>>
-    where C: std::borrow::Borrow<Cert>
-{
-    get_keys(certs, p, private_key_store, timestamp,
-             KeyType::KeyFlags(KeyFlags::empty().set_certification()),
-             options)
 }
 
 /// Returns the active certification, if any, for the specified bindings.
@@ -255,7 +79,7 @@ pub fn get_certification_keys<C>(certs: &[C], p: &dyn Policy,
 ///
 /// Note: if `n` User IDs are provided, then the returned vector has
 /// `n` elements.
-fn active_certification(config: &Config,
+fn active_certification(sq: &Sq,
                         cert: &Fingerprint, userids: Vec<UserID>,
                         issuer: &Key<openpgp::packet::key::PublicParts,
                                      openpgp::packet::key::UnspecifiedRole>)
@@ -263,7 +87,7 @@ fn active_certification(config: &Config,
 {
     // Look up the cert and find the certifications for the specified
     // User ID, if any.
-    let lc = config.cert_store_or_else()
+    let lc = sq.cert_store_or_else()
         .and_then(|cert_store| cert_store.lookup_by_cert_fpr(cert));
     let lc = match lc {
         Ok(lc) => lc,
@@ -299,14 +123,14 @@ fn active_certification(config: &Config,
             .iter()
             .filter(|sig| {
                 if let Some(ct) = sig.signature_creation_time() {
-                    ct <= config.time
+                    ct <= sq.time
                         && sig.signature_validity_period()
                         .map(|vp| {
-                            config.time < ct + vp
+                            sq.time < ct + vp
                         })
                         .unwrap_or(true)
                         && sig.get_issuers().iter().any(|i| i.aliases(&issuer_kh))
-                        && config.policy.signature(
+                        && sq.policy.signature(
                             sig, HashAlgoSecurity::CollisionResistance).is_ok()
                 } else {
                     false
@@ -323,7 +147,7 @@ fn active_certification(config: &Config,
         });
 
         // Return the first valid signature, which is the most recent one
-        // that is no younger than config.time.
+        // that is no younger than sq.time.
         let pk = ua.cert().primary_key().key();
         let certification = certifications.into_iter()
             .filter_map(|sig| {
@@ -396,40 +220,60 @@ pub fn cert_stub(cert: Cert,
     Ok(Cert::from_packets(packets.into_iter())?)
 }
 
-struct VHelper<'a, 'store> {
+pub struct VHelper<'c, 'store, 'rstore>
+    where 'store: 'rstore
+{
     #[allow(dead_code)]
-    config: &'a Config<'store>,
+    sq: &'c Sq<'store, 'rstore>,
     signatures: usize,
     certs: Option<Vec<Cert>>,
     labels: HashMap<KeyID, String>,
     trusted: HashSet<KeyID>,
+
+    /// Tracks the inner-most encryption container encountered.
+    sym_algo: Option<SymmetricAlgorithm>,
+    /// Tracks the inner-most encryption container encountered.
+    aead_algo: Option<AEADAlgorithm>,
+
+    // Tracks the signatures encountered.
     good_signatures: usize,
     good_checksums: usize,
     unknown_checksums: usize,
     bad_signatures: usize,
     bad_checksums: usize,
     broken_signatures: usize,
+    quiet: bool,
 }
 
-impl<'a, 'store> VHelper<'a, 'store> {
-    fn new(config: &'a Config<'store>, signatures: usize,
+impl<'c, 'store, 'rstore> VHelper<'c, 'store, 'rstore> {
+    fn new(sq: &'c Sq<'store, 'rstore>, signatures: usize,
            certs: Vec<Cert>)
            -> Self {
         VHelper {
-            config: config,
+            sq: sq,
             signatures,
             certs: Some(certs),
             labels: HashMap::new(),
             trusted: HashSet::new(),
+            sym_algo: None,
+            aead_algo: None,
             good_signatures: 0,
             good_checksums: 0,
             unknown_checksums: 0,
             bad_signatures: 0,
             bad_checksums: 0,
             broken_signatures: 0,
+            quiet: false,
         }
     }
 
+    /// Enables or disables quiet operation.
+    ///
+    /// In quiet operation, only errors are emitted.
+    pub fn quiet(&mut self, v: bool) {
+        self.quiet = v;
+    }
+
     fn print_status(&self) {
         fn p(s: &mut String, what: &str, quantity: usize) {
             if quantity > 0 {
@@ -456,10 +300,11 @@ impl<'a, 'store> VHelper<'a, 'store> {
     }
 
     fn print_sigs(&mut self, results: &[VerificationResult]) {
+        make_qprintln!(self.quiet);
         use crate::commands::pki::output::print_path;
         use crate::print_error_chain;
 
-        let reference_time = self.config.time;
+        let reference_time = self.sq.time;
 
         use self::VerificationError::*;
         for result in results {
@@ -480,6 +325,7 @@ impl<'a, 'store> VHelper<'a, 'store> {
                         n => format!("level {} notarizing checksum", n),
                     };
                     wprintln!("No key to check {} from {}", what, issuer);
+                    wprintln!("Consider running `sq network fetch {}`.", issuer);
                     self.unknown_checksums += 1;
                     continue;
                 },
@@ -521,23 +367,23 @@ impl<'a, 'store> VHelper<'a, 'store> {
             // Direct trust.
             let mut trusted = self.trusted.contains(&issuer);
             let mut prefix = "";
-            let trust_roots = self.config.trust_roots();
+            let trust_roots = self.sq.trust_roots();
             if ! trusted && ! trust_roots.is_empty() {
                 prefix = "  ";
 
                 // Web of trust.
-                wprintln!("Authenticating {} ({:?}) using the web of trust:",
+                qprintln!("Authenticating {} ({:?}) using the web of trust:",
                           cert_fpr, signer_userid);
 
-                if let Ok(Some(cert_store)) = self.config.cert_store() {
+                if let Ok(Some(cert_store)) = self.sq.cert_store() {
                     // Build the network.
                     let cert_store = sequoia_wot::store::CertStore::from_store(
-                        cert_store, &self.config.policy, reference_time);
+                        cert_store, self.sq.policy, reference_time);
 
                     let userids = if let Some(userid) = sig.signers_user_id() {
                         let userid = UserID::from(userid);
-                        wprintln!("{}Signature was made by {}",
-                                  prefix,
+                        wprintln!(indent=prefix,
+                                  "Signature was made by {}",
                                   String::from_utf8_lossy(userid.value()));
                         vec![ userid ]
                     } else {
@@ -545,47 +391,47 @@ impl<'a, 'store> VHelper<'a, 'store> {
                     };
 
                     if userids.is_empty() {
-                        wprintln!("{}{} cannot be authenticated.  \
+                        wprintln!(indent=prefix,
+                                  "{} cannot be authenticated.  \
                                    It has no User IDs",
-                                  prefix, cert_fpr);
-                    } else if let Ok(n) = sequoia_wot::Network::new(&cert_store) {
-                        let mut q = sequoia_wot::QueryBuilder::new(&n);
-                        q.roots(sequoia_wot::Roots::new(trust_roots.into_iter()));
-                        let q = q.build();
+                                  cert_fpr);
+                    } else {
+                        let n = sequoia_wot::NetworkBuilder::rooted(
+                            &cert_store, &*trust_roots).build();
 
                         let authenticated_userids
                             = userids.into_iter().filter(|userid| {
                                 let userid_str =
                                     String::from_utf8_lossy(userid.value());
 
-                                let paths = q.authenticate(
+                                let paths = n.authenticate(
                                     userid, cert.fingerprint(),
-                                    // XXX: Make this user configurable.
+                                    // XXX: Make this user squrable.
                                     sequoia_wot::FULLY_TRUSTED);
 
                                 let amount = paths.amount();
                                 let authenticated = if amount >= sequoia_wot::FULLY_TRUSTED {
-                                    wprintln!("{}Fully authenticated \
+                                    wprintln!(indent=prefix,
+                                              "Fully authenticated \
                                                ({} of {}) {}, {}",
-                                              prefix,
                                               amount,
                                               sequoia_wot::FULLY_TRUSTED,
                                               cert_fpr,
                                               userid_str);
                                     true
                                 } else if amount > 0 {
-                                    wprintln!("{}Partially authenticated \
+                                    wprintln!(indent=prefix,
+                                              "Partially authenticated \
                                                ({} of {}) {}, {:?} ",
-                                              prefix,
                                               amount,
                                               sequoia_wot::FULLY_TRUSTED,
                                               cert_fpr,
                                               userid_str);
                                     false
                                 } else {
-                                    wprintln!("{}{}: {:?} is unauthenticated \
+                                    wprintln!(indent=prefix,
+                                              "{}: {:?} is unauthenticated \
                                                and may be an impersonation!",
-                                              prefix,
                                               cert_fpr,
                                               userid_str);
                                     false
@@ -593,7 +439,7 @@ impl<'a, 'store> VHelper<'a, 'store> {
 
                                 for (i, (path, amount)) in paths.iter().enumerate() {
                                     let prefix = if paths.len() > 1 {
-                                        wprintln!("{}  Path #{} of {}, \
+                                        qprintln!("{}  Path #{} of {}, \
                                                   trust amount {}:",
                                                  prefix,
                                                  i + 1, paths.len(), amount);
@@ -602,7 +448,11 @@ impl<'a, 'store> VHelper<'a, 'store> {
                                         format!("{}  ", prefix)
                                     };
 
-                                    print_path(&path.into(), userid, &prefix)
+                                    if ! self.quiet {
+                                        let _ =
+                                            print_path(&path.into(), userid,
+                                                       &prefix);
+                                    }
                                 }
 
                                 authenticated
@@ -616,11 +466,9 @@ impl<'a, 'store> VHelper<'a, 'store> {
                             signer_userid = String::from_utf8_lossy(
                                 authenticated_userids[0].value()).to_string();
                         }
-                    } else {
-                        wprintln!("Failed to build web of trust network.");
                     }
                 } else {
-                    wprintln!("Skipping, certificate store has been disabled");
+                    qprintln!("Skipping, certificate store has been disabled");
                 }
             }
 
@@ -630,31 +478,41 @@ impl<'a, 'store> VHelper<'a, 'store> {
             let level = sig.level();
             match (level == 0, trusted) {
                 (true,  true)  => {
-                    wprintln!("{}Good signature from {} ({:?})",
-                              prefix, label, signer_userid);
+                    wprintln!(indent=prefix,
+                              "Good signature from {} ({:?})",
+                              label, signer_userid);
                 }
                 (false, true)  => {
-                    wprintln!("{}Good level {} notarization from {} ({:?})",
-                              prefix, level, label, signer_userid);
+                    wprintln!(indent=prefix,
+                              "Good level {} notarization from {} ({:?})",
+                              level, label, signer_userid);
                 }
                 (true,  false) => {
-                    wprintln!("{}Unauthenticated checksum from {} ({:?})",
-                              prefix, label, signer_userid);
-                    wprintln!("{}  After checking that {} belongs to {:?}, \
-                               you can authenticate the binding using \
-                               'sq link add {} {:?}'.",
-                              prefix, issuer_str, signer_userid,
+                    wprintln!(indent=prefix,
+                              "Unauthenticated checksum from {} ({:?})",
+                              label, signer_userid);
+                    eprintln!();
+                    wprintln!(indent=prefix,
+                              "After checking that {} belongs to {:?}, \
+                               you can authenticate the binding using:",
                               issuer_str, signer_userid);
+                    eprintln!();
+                    eprintln!("{}  $ sq pki link add {} {:?}",
+                              prefix, issuer_str, signer_userid);
                 }
                 (false, false) => {
-                    wprintln!("{}Unauthenticated level {} notarizing \
+                    wprintln!(indent=prefix,
+                              "Unauthenticated level {} notarizing \
                                checksum from {} ({:?})",
-                              prefix, level, label, signer_userid);
-                    wprintln!("{}  After checking that {} belongs to {:?}, \
-                               you can authenticate the binding using \
-                               'sq link add {} {:?}'.",
-                              prefix, issuer_str, signer_userid,
+                              level, label, signer_userid);
+                    eprintln!();
+                    wprintln!(indent=prefix,
+                              "After checking that {} belongs to {:?}, \
+                               you can authenticate the binding using:",
                               issuer_str, signer_userid);
+                    eprintln!();
+                    eprintln!("{}  $ sq pki link add {} {:?}",
+                              prefix, issuer_str, signer_userid);
                 }
             };
 
@@ -664,12 +522,14 @@ impl<'a, 'store> VHelper<'a, 'store> {
                 self.good_checksums += 1;
             }
 
-            wprintln!("");
+            qprintln!("");
         }
     }
 }
 
-impl<'a, 'store> VerificationHelper for VHelper<'a, 'store> {
+impl<'c, 'store, 'rstore> VerificationHelper for VHelper<'c, 'store, 'rstore>
+    where 'store: 'rstore
+{
     fn get_certs(&mut self, ids: &[openpgp::KeyHandle]) -> Result<Vec<Cert>> {
         let mut certs = BTreeMap::new();
 
@@ -699,7 +559,7 @@ impl<'a, 'store> VerificationHelper for VHelper<'a, 'store> {
         // Avoid initializing the certificate store if we don't actually
         // need to.
         if ! ids.is_empty() {
-            if let Ok(Some(cert_store)) = self.config.cert_store() {
+            if let Ok(Some(cert_store)) = self.sq.cert_store() {
                 for id in ids.iter() {
                     for c in cert_store.lookup_by_cert_or_subkey(id)
                         .unwrap_or_default()
@@ -723,17 +583,22 @@ impl<'a, 'store> VerificationHelper for VHelper<'a, 'store> {
     }
 
     fn check(&mut self, structure: MessageStructure) -> Result<()> {
+        make_qprintln!(self.quiet);
         for layer in structure {
             match layer {
                 MessageLayer::Compression { algo } =>
-                    wprintln!("Compressed using {}", algo),
-                MessageLayer::Encryption { sym_algo, aead_algo } =>
+                    qprintln!("Compressed using {}", algo),
+                MessageLayer::Encryption { sym_algo, aead_algo } => {
+                    self.sym_algo = Some(sym_algo);
+                    self.aead_algo = aead_algo;
+
                     if let Some(aead_algo) = aead_algo {
-                        wprintln!("Encrypted and protected using {}/{}",
+                        qprintln!("Encrypted and protected using {}/{}",
                                   sym_algo, aead_algo);
                     } else {
-                        wprintln!("Encrypted using {}", sym_algo);
-                    },
+                        qprintln!("Encrypted using {}", sym_algo);
+                    }
+                },
                 MessageLayer::SignatureGroup { ref results } =>
                     self.print_sigs(results),
             }
@@ -743,7 +608,9 @@ impl<'a, 'store> VerificationHelper for VHelper<'a, 'store> {
             && self.bad_signatures + self.bad_checksums == 0 {
             Ok(())
         } else {
-            self.print_status();
+            if ! self.quiet {
+                self.print_status();
+            }
             Err(anyhow::anyhow!("Verification failed: could not fully \
                                  authenticate any signatures"))
         }
