@@ -2,12 +2,10 @@ use anyhow::Context as _;
 use std::collections::HashMap;
 use std::io;
 
-use terminal_size::terminal_size;
-
-use sequoia_net::pks;
 use sequoia_openpgp as openpgp;
 use openpgp::types::SymmetricAlgorithm;
 use openpgp::fmt::hex;
+use openpgp::KeyHandle;
 use openpgp::crypto::{self, SessionKey, Decryptor, Password};
 use openpgp::{Fingerprint, Cert, KeyID, Result};
 use openpgp::packet;
@@ -20,23 +18,29 @@ use openpgp::parse::{
 use openpgp::parse::stream::{
     VerificationHelper, DecryptionHelper, DecryptorBuilder, MessageStructure,
 };
+use sequoia_openpgp::types::KeyFlags;
+
+use sequoia_cert_store as cert_store;
+use cert_store::store::StoreError;
+
+use sequoia_keystore as keystore;
 
 use crate::{
     cli,
     commands::{
-        toolbox::packet::dump::PacketDumper,
         VHelper,
     },
     common::password,
-    Config,
+    Sq,
     load_certs,
+    load_keys,
 };
 
-pub fn dispatch(config: Config, command: cli::decrypt::Command) -> Result<()> {
+pub fn dispatch(sq: Sq, command: cli::decrypt::Command) -> Result<()> {
     tracer!(TRACE, "decrypt::dispatch");
 
     let mut input = command.input.open()?;
-    let mut output = command.output.create_safe(config.force)?;
+    let mut output = command.output.create_safe(sq.force)?;
 
     let certs = load_certs(
         command.sender_cert_file.iter().map(|s| s.as_ref()),
@@ -55,17 +59,13 @@ pub fn dispatch(config: Config, command: cli::decrypt::Command) -> Result<()> {
             1
         }
     });
-    // TODO: should this be load_keys?
     let secrets =
-        load_certs(command.secret_key_file.iter().map(|s| s.as_ref()))?;
-    let private_key_store = command.private_key_store;
+        load_keys(command.secret_key_file.iter().map(|s| s.as_ref()))?;
     let session_keys = command.session_key;
-    decrypt(config, private_key_store.as_deref(),
-            &mut input, &mut output,
+    decrypt(sq, &mut input, &mut output,
             signatures, certs, secrets,
             command.dump_session_key,
-            session_keys,
-            command.dump, command.hex)?;
+            session_keys)?;
 
     Ok(())
 }
@@ -106,53 +106,47 @@ impl PrivateKey for LocalPrivateKey {
     }
 }
 
-struct RemotePrivateKey {
-    key: Key<key::PublicParts, key::UnspecifiedRole>,
-    store: String,
-}
-
-impl RemotePrivateKey {
-    fn new(key: Key<key::PublicParts, key::UnspecifiedRole>, store: String) -> Self {
-        Self {
-            key,
-            store,
-        }
-    }
-}
-
-impl PrivateKey for RemotePrivateKey {
-    fn get_unlocked(&self) -> Option<Box<dyn Decryptor>> {
-        // getting already unlocked keys is not implemented right now
-        None
-    }
-
-    fn unlock(&mut self, p: &Password) -> Result<Box<dyn Decryptor>> {
-        Ok(pks::unlock_decryptor(&self.store, self.key.clone(), p)?)
-    }
-}
-
-struct Helper<'a, 'certdb> {
-    vhelper: VHelper<'a, 'certdb>,
+pub struct Helper<'c, 'store, 'rstore>
+    where 'store: 'rstore
+{
+    vhelper: VHelper<'c, 'store, 'rstore>,
     secret_keys: HashMap<KeyID, Box<dyn PrivateKey>>,
     key_identities: HashMap<KeyID, Fingerprint>,
     key_hints: HashMap<KeyID, String>,
     session_keys: Vec<cli::types::SessionKey>,
     dump_session_key: bool,
-    dumper: Option<PacketDumper>,
 }
 
-impl<'a, 'certdb> Helper<'a, 'certdb> {
-    fn new(config: &'a Config<'certdb>, private_key_store: Option<&str>,
-           signatures: usize, certs: Vec<Cert>, secrets: Vec<Cert>,
-           session_keys: Vec<cli::types::SessionKey>,
-           dump_session_key: bool, dump: bool)
-           -> Self
+impl<'c, 'store, 'rstore> std::ops::Deref for Helper<'c, 'store, 'rstore>
+    where 'store: 'rstore
+{
+    type Target = VHelper<'c, 'store, 'rstore>;
+
+    fn deref(&self) -> &Self::Target {
+        &self.vhelper
+    }
+}
+
+impl<'c, 'store, 'rstore> std::ops::DerefMut for Helper<'c, 'store, 'rstore> {
+    fn deref_mut(&mut self) -> &mut Self::Target {
+        &mut self.vhelper
+    }
+}
+
+impl<'c, 'store, 'rstore> Helper<'c, 'store, 'rstore>
+    where 'store: 'rstore
+{
+    pub fn new(sq: &'c Sq<'store, 'rstore>,
+               signatures: usize, certs: Vec<Cert>, secrets: Vec<Cert>,
+               session_keys: Vec<cli::types::SessionKey>,
+               dump_session_key: bool)
+               -> Self
     {
         let mut keys: HashMap<KeyID, Box<dyn PrivateKey>> = HashMap::new();
         let mut identities: HashMap<KeyID, Fingerprint> = HashMap::new();
         let mut hints: HashMap<KeyID, String> = HashMap::new();
         for tsk in secrets {
-            let hint = match tsk.with_policy(&config.policy, None)
+            let hint = match tsk.with_policy(sq.policy, None)
                 .and_then(|valid_cert| valid_cert.primary_userid()).ok()
             {
                 Some(uid) => format!("{} ({})", uid.userid(),
@@ -162,7 +156,7 @@ impl<'a, 'certdb> Helper<'a, 'certdb> {
 
             for ka in tsk.keys()
             // XXX: Should use the message's creation time that we do not know.
-                .with_policy(&config.policy, None)
+                .with_policy(sq.policy, None)
                 .for_transport_encryption().for_storage_encryption()
             {
                 let id: KeyID = ka.key().fingerprint().into();
@@ -170,8 +164,6 @@ impl<'a, 'certdb> Helper<'a, 'certdb> {
                 keys.insert(id.clone(),
                     if let Ok(key) = key.parts_as_secret() {
                         Box::new(LocalPrivateKey::new(key.clone()))
-                    } else if let Some(store) = private_key_store {
-                        Box::new(RemotePrivateKey::new(key.clone(), store.to_string()))
                     } else {
                         panic!("Cert does not contain secret keys and private-key-store option has not been set.");
                     }
@@ -182,20 +174,30 @@ impl<'a, 'certdb> Helper<'a, 'certdb> {
         }
 
         Helper {
-            vhelper: VHelper::new(config, signatures, certs),
+            vhelper: VHelper::new(sq, signatures, certs),
             secret_keys: keys,
             key_identities: identities,
             key_hints: hints,
             session_keys,
             dump_session_key,
-            dumper: if dump {
-                let width = terminal_size()
-                    .map(|(width, _height)| width.0.into())
-                    .unwrap_or(80);
-                Some(PacketDumper::new(width, false))
-            } else {
-                None
-            },
+        }
+    }
+
+    /// Checks if a session key can decrypt the packet parser using
+    /// `decrypt`.
+    fn try_session_key<D>(&self, keyid: &KeyID,
+                          algo: SymmetricAlgorithm, sk: SessionKey,
+                          decrypt: &mut D)
+        -> Option<Option<Fingerprint>>
+        where D: FnMut(SymmetricAlgorithm, &SessionKey) -> bool
+    {
+        if decrypt(algo, &sk) {
+            if self.dump_session_key {
+                wprintln!("Session key: {}", hex::encode(&sk));
+            }
+            Some(self.key_identities.get(keyid).cloned())
+        } else {
+            None
         }
     }
 
@@ -209,33 +211,14 @@ impl<'a, 'certdb> Helper<'a, 'certdb> {
         where D: FnMut(SymmetricAlgorithm, &SessionKey) -> bool
     {
         let keyid = keypair.public().fingerprint().into();
-        match pkesk.decrypt(&mut *keypair, sym_algo)
-            .and_then(|(algo, sk)| {
-                if decrypt(algo, &sk) { Some(sk) } else { None }
-            })
-        {
-            Some(sk) => {
-                if self.dump_session_key {
-                    wprintln!("Session key: {}", hex::encode(&sk));
-                }
-                Some(self.key_identities.get(&keyid).cloned())
-            },
-            None => None,
-        }
+        let (sym_algo, sk) = pkesk.decrypt(&mut *keypair, sym_algo)?;
+        self.try_session_key(&keyid, sym_algo, sk, decrypt)
     }
 }
 
-impl<'a, 'certdb> VerificationHelper for Helper<'a, 'certdb> {
-    fn inspect(&mut self, pp: &PacketParser) -> Result<()> {
-        if let Some(dumper) = self.dumper.as_mut() {
-            dumper.packet(&mut io::stderr(),
-                          pp.recursion_depth() as usize,
-                          pp.header().clone(), pp.packet.clone(),
-                          pp.map().cloned(), Vec::new())?;
-        }
-        Ok(())
-    }
-
+impl<'c, 'store, 'rstore> VerificationHelper for Helper<'c, 'store, 'rstore>
+    where 'store: 'rstore
+{
     fn get_certs(&mut self, ids: &[openpgp::KeyHandle]) -> Result<Vec<Cert>> {
         self.vhelper.get_certs(ids)
     }
@@ -244,12 +227,16 @@ impl<'a, 'certdb> VerificationHelper for Helper<'a, 'certdb> {
     }
 }
 
-impl<'a, 'certdb> DecryptionHelper for Helper<'a, 'certdb> {
+impl<'c, 'store, 'rstore> DecryptionHelper for Helper<'c, 'store, 'rstore>
+    where 'store: 'rstore
+{
     fn decrypt<D>(&mut self, pkesks: &[PKESK], skesks: &[SKESK],
                   sym_algo: Option<SymmetricAlgorithm>,
                   mut decrypt: D) -> openpgp::Result<Option<Fingerprint>>
         where D: FnMut(SymmetricAlgorithm, &SessionKey) -> bool
     {
+        make_qprintln!(self.quiet);
+
         // Before anything else, try the session keys
         for sk in &self.session_keys {
             let decrypted = if let Some(sa) = sk.symmetric_algo {
@@ -262,11 +249,15 @@ impl<'a, 'certdb> DecryptionHelper for Helper<'a, 'certdb> {
                     .any(|sa| decrypt(sa, &sk.session_key))
             };
             if decrypted {
-                wprintln!("Encrypted with Session Key {}", sk.display_sensitive());
+                qprintln!("Encrypted with Session Key {}",
+                          sk.display_sensitive());
                 return Ok(None);
             }
         }
 
+        // Now, we try the secret keys that the user supplied on the
+        // command line.
+
         // First, we try those keys that we can use without prompting
         // for a password.
         for pkesk in pkesks {
@@ -371,12 +362,116 @@ impl<'a, 'certdb> DecryptionHelper for Helper<'a, 'certdb> {
             }
         }
 
+        // Try the key store.
+        match self.vhelper.sq.key_store_or_else() {
+            Ok(ks) => {
+                let mut ks = ks.lock().unwrap();
+                match ks.decrypt(&pkesks[..]) {
+                    // Success!
+                    Ok((_i, fpr, sym_algo, sk)) => {
+                        if let Some(fp) =
+                            self.try_session_key(
+                                &KeyID::from(fpr), sym_algo, sk, &mut decrypt)
+                        {
+                            return Ok(fp);
+                        }
+                    }
+
+                    Err(err) => {
+                        match err.downcast() {
+                            Ok(keystore::Error::InaccessibleDecryptionKey(keys)) => {
+                                for key_status in keys.into_iter() {
+                                    let pkesk = key_status.pkesk().clone();
+                                    let mut key = key_status.into_key();
+                                    let keyid = key.keyid();
+                                    let userid = self.sq.best_userid_for(
+                                        &KeyHandle::from(&keyid), true);
+
+                                    let keypair = loop {
+                                        let password = rpassword::prompt_password(
+                                            format!(
+                                                "Enter password to unlock {}, {}: ",
+                                                keyid, userid))?
+                                            .into();
+
+                                        if let Ok(()) = key.unlock(password) {
+                                            break Box::new(key);
+                                        } else {
+                                            wprintln!("Bad password.");
+                                        }
+                                    };
+
+                                    if let Some(fp) = self.try_decrypt(
+                                        &pkesk, sym_algo, keypair, &mut decrypt)
+                                    {
+                                        return Ok(fp);
+                                    }
+                                }
+                            }
+                            // Failed to decrypt using the keystore.
+                            Ok(_err) => (),
+                            Err(_err) => (),
+                        }
+                    }
+                }
+            }
+            Err(err) => {
+                wprintln!("Warning: unable to connect to keystore: {}",
+                          err);
+            }
+        }
+
         if skesks.is_empty() {
+            let recipients = pkesks.iter()
+                .filter_map(|p| {
+                    let recipient = p.recipient();
+                    if recipient.is_wildcard() {
+                        None
+                    } else {
+                        Some(recipient)
+                    }
+                });
+            wprintln!("No key to decrypt message.  The message appears \
+                       to be encrypted to:");
+            eprintln!();
+            for recipient in recipients.into_iter() {
+                let certs = self.sq.lookup(
+                    std::iter::once(KeyHandle::from(recipient)),
+                    Some(KeyFlags::empty()
+                         .set_storage_encryption()
+                         .set_transport_encryption()),
+                    false,
+                    true);
+
+                match certs {
+                    Ok(certs) => {
+                        for cert in certs {
+                            eprintln!("  - {}, {}",
+                                      cert.fingerprint(),
+                                      self.sq.best_userid(&cert, true));
+                        }
+                    }
+                    Err(err) => {
+                        if let Some(StoreError::NotFound(_))
+                            = err.downcast_ref()
+                        {
+                            eprintln!("  - {}, certificate not found",
+                                      recipient);
+                        } else {
+                            eprintln!("  - {}, error looking up certificate: {}",
+                                      recipient, err);
+                        }
+                    }
+                };
+            }
+            eprintln!();
+
             return
                 Err(anyhow::anyhow!("No key to decrypt message"));
         }
 
         // Finally, try to decrypt using the SKESKs.
+        let mut first = true;
         loop {
             let password = password::prompt_to_unlock("message")?;
 
@@ -391,40 +486,44 @@ impl<'a, 'certdb> DecryptionHelper for Helper<'a, 'certdb> {
                 }
             }
 
-            wprintln!("Incorrect password.");
+            if password.map(|p| p.is_empty()) {
+                break Err(anyhow::anyhow!("Decryption failed."));
+            }
+
+            if first {
+                wprintln!("Incorrect password.  \
+                           Hint: enter empty password to cancel.");
+                first = false;
+            } else {
+                wprintln!("Incorrect password.");
+            }
         }
     }
 }
 
 // Allow too many arguments now, should be reworked later
 #[allow(clippy::too_many_arguments)]
-pub fn decrypt(config: Config,
-               private_key_store: Option<&str>,
+pub fn decrypt(sq: Sq,
                input: &mut (dyn io::Read + Sync + Send),
                output: &mut dyn io::Write,
                signatures: usize, certs: Vec<Cert>, secrets: Vec<Cert>,
                dump_session_key: bool,
-               sk: Vec<cli::types::SessionKey>,
-               dump: bool, hex: bool)
+               sk: Vec<cli::types::SessionKey>)
                -> Result<()> {
-    let helper = Helper::new(&config, private_key_store, signatures, certs,
-                             secrets, sk, dump_session_key, dump || hex);
+    let helper = Helper::new(&sq, signatures, certs,
+                             secrets, sk, dump_session_key);
     let mut decryptor = DecryptorBuilder::from_reader(input)?
-        .mapping(hex)
-        .with_policy(&config.policy, None, helper)
+        .with_policy(sq.policy, None, helper)
         .context("Decryption failed")?;
 
     io::copy(&mut decryptor, output).context("Decryption failed")?;
 
     let helper = decryptor.into_helper();
-    if let Some(dumper) = helper.dumper.as_ref() {
-        dumper.flush(&mut io::stderr())?;
-    }
     helper.vhelper.print_status();
     Ok(())
 }
 
-pub fn decrypt_unwrap(config: Config,
+pub fn decrypt_unwrap(sq: Sq,
                       input: &mut (dyn io::Read + Sync + Send),
                       output: &mut dyn io::Write,
                       secrets: Vec<Cert>,
@@ -432,9 +531,9 @@ pub fn decrypt_unwrap(config: Config,
                       dump_session_key: bool)
                       -> Result<()>
 {
-    let mut helper = Helper::new(&config, None, 0, Vec::new(), secrets,
+    let mut helper = Helper::new(&sq, 0, Vec::new(), secrets,
                                  session_keys,
-                                 dump_session_key, false);
+                                 dump_session_key);
 
     let mut ppr = PacketParser::from_reader(input)?;
 
