@@ -4,56 +4,51 @@ use std::time::{Duration, SystemTime};
 
 use anyhow::Context;
 
+use buffered_reader::BufferedReader;
+
 use sequoia_openpgp as openpgp;
 use openpgp::{KeyHandle, Packet, Result};
+use openpgp::armor::ReaderMode;
 use openpgp::cert::prelude::*;
 use openpgp::packet::{
     Signature,
     key::PublicParts,
 };
-use openpgp::parse::{Parse, PacketParserResult};
+use openpgp::parse::{Dearmor, Parse, PacketParserBuilder, PacketParserResult};
 use openpgp::policy::{Policy, HashAlgoSecurity};
 use openpgp::packet::key::SecretKeyMaterial;
-use openpgp::types::SignatureType;
+use openpgp::types::{
+    ReasonForRevocation,
+    SignatureType,
+};
 
 use sequoia_cert_store as cert_store;
 use cert_store::Store;
 
 use crate::Convert;
 
-use crate::Config;
+use crate::Sq;
+use crate::one_line_error_chain;
 use crate::SECONDS_IN_YEAR;
 use crate::SECONDS_IN_DAY;
 
 use crate::cli::inspect;
 use crate::cli::types::FileOrStdout;
 
-pub fn dispatch(config: Config, c: inspect::Command)
+pub fn dispatch(mut sq: Sq, c: inspect::Command)
     -> Result<()>
 {
     // sq inspect does not have --output, but commands::inspect does.
     // Work around this mismatch by always creating a stdout output.
     let output_type = FileOrStdout::default();
-    let output = &mut output_type.create_unsafe(config.force)?;
-
-    let policy = &config.policy;
-    let time = Some(config.time);
+    let output = &mut output_type.create_unsafe(sq.force)?;
 
     let print_certifications = c.certifications;
 
     let input = c.input;
-    write!(output, "{}: ", input)?;
-
-    let mut type_called = false;  // Did we print the type yet?
-    let mut encrypted = false;    // Is it an encrypted message?
-    let mut packets = Vec::new(); // Accumulator for packets.
-    let mut pkesks = Vec::new();  // Accumulator for PKESKs.
-    let mut n_skesks = 0;         // Number of SKESKs.
-    let mut sigs = Vec::new();    // Accumulator for signatures.
-    let mut literal_prefix = Vec::new();
 
     let mut bytes: Vec<u8> = Vec::new();
-    let mut ppr = if c.cert.is_empty() {
+    if c.cert.is_empty() {
         if let Some(path) = input.inner() {
             if ! path.exists() &&
                 format!("{}", input).parse::<KeyHandle>().is_ok() {
@@ -63,9 +58,11 @@ pub fn dispatch(config: Config, c: inspect::Command)
             }
         }
 
-        openpgp::parse::PacketParser::from_reader(input.open()?)?
+        inspect(&mut sq, input.open()?,
+                Some(&input.to_string()), output,
+                print_certifications)
     } else {
-        let cert_store = config.cert_store_or_else()?;
+        let cert_store = sq.cert_store_or_else()?;
         for cert in c.cert.into_iter() {
             let certs = cert_store.lookup_by_cert_or_subkey(&cert)
                 .with_context(|| format!("Looking up {}", cert))?;
@@ -77,8 +74,45 @@ pub fn dispatch(config: Config, c: inspect::Command)
             }
         }
 
-        openpgp::parse::PacketParser::from_bytes(&bytes)?
-    };
+        let br = buffered_reader::Memory::with_cookie(
+            &bytes, sequoia_openpgp::parse::Cookie::default());
+        inspect(&mut sq, br, None, output,
+                print_certifications)
+    }
+}
+
+/// Inspects OpenPGP data.
+///
+/// The data is read from `input`.  `input_filename` is the name of
+/// the file, if available.  This is only used for display purposes.
+/// The output is written to `output`.
+///
+/// If `print_certifications` is set, also shows information about
+/// certifications.
+pub fn inspect<'a, R>(sq: &mut Sq,
+                      input: R,
+                      input_filename: Option<&str>,
+                      output: &mut Box<dyn std::io::Write + Send + Sync>,
+                      print_certifications: bool)
+    -> Result<()>
+where R: BufferedReader<sequoia_openpgp::parse::Cookie> + 'a,
+{
+    let time = Some(sq.time);
+
+    let mut ppr = openpgp::parse::PacketParser::from_buffered_reader(input)?;
+
+  loop {
+    if let Some(input_filename) = input_filename {
+        write!(output, "{}: ", input_filename)?;
+    }
+
+    let mut type_called = false;  // Did we print the type yet?
+    let mut encrypted = false;    // Is it an encrypted message?
+    let mut packets = Vec::new(); // Accumulator for packets.
+    let mut pkesks = Vec::new();  // Accumulator for PKESKs.
+    let mut n_skesks = 0;         // Number of SKESKs.
+    let mut sigs = Vec::new();    // Accumulator for signatures.
+    let mut literal_prefix = Vec::new();
 
     while let PacketParserResult::Some(mut pp) = ppr {
         match pp.packet {
@@ -95,7 +129,8 @@ pub fn dispatch(config: Config, c: inspect::Command)
                         std::mem::take(&mut packets));
                     let cert = openpgp::Cert::try_from(pp)?;
                     inspect_cert(
-                        policy,
+                        sq,
+                        sq.policy,
                         time,
                         output,
                         &cert,
@@ -133,6 +168,17 @@ pub fn dispatch(config: Config, c: inspect::Command)
         let is_cert = eof.is_cert();
         let is_keyring = eof.is_keyring();
 
+        // Now, the parser is exhausted, but we may find another
+        // armored blob.  Note that this can only happen if the first
+        // set of packets was also armored.
+        let next_ppr =
+            PacketParserBuilder::from_buffered_reader(eof.into_reader())
+            .and_then(
+                |builder| builder
+                    .dearmor(Dearmor::Enabled(
+                        ReaderMode::Tolerant(None)))
+                    .build());
+
         if is_message.is_ok() {
             writeln!(output, "{}OpenPGP Message.",
                      match (encrypted, ! sigs.is_empty()) {
@@ -148,7 +194,7 @@ pub fn dispatch(config: Config, c: inspect::Command)
             for pkesk in pkesks.iter() {
                 writeln!(output, "      Recipient: {}", pkesk.recipient())?;
             }
-            inspect_signatures(output, &sigs)?;
+            inspect_signatures(sq, output, &sigs)?;
             if ! literal_prefix.is_empty() {
                 writeln!(output, "           Data: {:?}{}",
                          String::from_utf8_lossy(&literal_prefix),
@@ -158,14 +204,15 @@ pub fn dispatch(config: Config, c: inspect::Command)
         } else if is_cert.is_ok() || is_keyring.is_ok() {
             let pp = openpgp::PacketPile::from(packets);
             let cert = openpgp::Cert::try_from(pp)?;
-            inspect_cert(policy, time, output, &cert, print_certifications)?;
+            inspect_cert(sq, sq.policy, time, output, &cert,
+                         print_certifications)?;
         } else if packets.is_empty() && ! sigs.is_empty() {
             if sigs.iter().all(is_revocation_sig) {
                 writeln!(output, "Revocation Certificate{}.",
                          if sigs.len() > 1 { "s" } else { "" })?;
                 writeln!(output)?;
                 for sig in sigs {
-                    inspect_bare_revocation(output, &sig)?;
+                    inspect_bare_revocation(sq, output, &sig)?;
                 }
                 writeln!(output, "           Note: \
                                   Signatures have NOT been verified!")?;
@@ -173,21 +220,46 @@ pub fn dispatch(config: Config, c: inspect::Command)
                 writeln!(output, "Detached signature{}.",
                          if sigs.len() > 1 { "s" } else { "" })?;
                 writeln!(output)?;
-                inspect_signatures(output, &sigs)?;
+                inspect_signatures(sq, output, &sigs)?;
             }
         } else if packets.is_empty() {
             writeln!(output, "No OpenPGP data.")?;
         } else {
             writeln!(output, "Unknown sequence of OpenPGP packets.")?;
-            writeln!(output, "  Message: {}", is_message.unwrap_err())?;
-            writeln!(output, "  Cert: {}", is_cert.unwrap_err())?;
-            writeln!(output, "  Keyring: {}", is_keyring.unwrap_err())?;
+            writeln!(output, "  Message: {}", is_message.as_ref().unwrap_err())?;
+            writeln!(output, "  Cert: {}", is_cert.as_ref().unwrap_err())?;
+            writeln!(output, "  Keyring: {}", is_keyring.as_ref().unwrap_err())?;
             writeln!(output)?;
-            writeln!(output, "Hint: Try 'sq packet dump {}'", input)?;
+            if let Some(filename) = input_filename {
+                writeln!(output, "Hint: Try 'sq toolbox packet dump {}'",
+                         filename)?;
+            }
+        }
+
+        // See if there is another armor block.
+        match next_ppr {
+            Ok(ppr_) => {
+                writeln!(output, "Note: There is another block of armored \
+                                  OpenPGP data.")?;
+
+                if is_message.is_ok() {
+                    writeln!(output, "Note: Data concatenated to a message is \
+                                      likely an error.")?;
+                } else if is_cert.is_ok() || is_keyring.is_ok() {
+                    writeln!(output, "Note: This is a non-standard extension \
+                                      to OpenPGP.")?;
+                }
+                writeln!(output)?;
+
+                ppr = ppr_;
+                continue;
+            },
+            Err(_) => break,
         }
     } else {
         unreachable!()
     }
+  }
 
     Ok(())
 }
@@ -213,6 +285,7 @@ fn is_revocation_cert(c: &Cert) -> bool {
 }
 
 fn inspect_cert(
+    sq: &mut Sq,
     policy: &dyn Policy,
     time: Option<SystemTime>,
     output: &mut dyn io::Write,
@@ -230,6 +303,7 @@ fn inspect_cert(
     writeln!(output, "    Fingerprint: {}", cert.fingerprint())?;
     inspect_revocation(output, "", cert.revocation_status(policy, None))?;
     inspect_key(
+        sq,
         policy,
         time,
         output,
@@ -251,6 +325,7 @@ fn inspect_cert(
             Err(e) => print_error_chain(output, &e)?,
         }
         inspect_key(
+            sq,
             policy,
             time,
             output,
@@ -281,7 +356,7 @@ fn inspect_cert(
             }
             Err(e) => print_error_chain(output, &e)?,
         }
-        inspect_certifications(output, policy,
+        inspect_certifications(sq, output, policy,
                                uidb.certifications(),
                                print_certifications)?;
         writeln!(output)?;
@@ -299,7 +374,7 @@ fn inspect_cert(
             }
             Err(e) => print_error_chain(output, &e)?,
         }
-        inspect_certifications(output, policy,
+        inspect_certifications(sq, output, policy,
                                uab.certifications(),
                                print_certifications)?;
         writeln!(output)?;
@@ -315,7 +390,7 @@ fn inspect_cert(
             }
             Err(e) => print_error_chain(output, &e)?,
         }
-        inspect_certifications(output, policy,
+        inspect_certifications(sq, output, policy,
                                ub.certifications(),
                                print_certifications)?;
         writeln!(output)?;
@@ -329,6 +404,7 @@ fn inspect_cert(
 }
 
 fn inspect_key(
+    sq: &mut Sq,
     policy: &dyn Policy,
     time: Option<SystemTime>,
     output: &mut dyn io::Write,
@@ -341,12 +417,16 @@ fn inspect_key(
     let vka = match ka.with_policy(policy, time) {
         Ok(vka) => {
             if let Err(e) = vka.alive() {
-                writeln!(output, "{}                 Invalid: {}", indent, e)?;
+                writeln!(output, "{}                 Invalid: {}",
+                         indent,
+                         one_line_error_chain(&e))?;
             }
             Some(vka)
         },
         Err(e) => {
-            writeln!(output, "{}                 Invalid: {}", indent, e)?;
+            writeln!(output, "{}                 Invalid: {}",
+                     indent,
+                     one_line_error_chain(&e))?;
             None
         },
     };
@@ -379,7 +459,7 @@ fn inspect_key(
             writeln!(output, "{}      Key flags: {}", indent, flags)?;
         }
     }
-    inspect_certifications(output, policy,
+    inspect_certifications(sq, output, policy,
                            bundle.certifications().iter(),
                            print_certifications)?;
 
@@ -391,35 +471,28 @@ fn print_reasons(output: &mut dyn io::Write, indent: &str,
                  third_party: bool, sigs: &[&Signature])
                  -> Result<()> {
     for sig in sigs {
-        if let Some((r, msg)) = sig.reason_for_revocation() {
-            writeln!(output, "{}                  - {}", indent, r)?;
-            if third_party {
-                writeln!(output, "{}                    Issued by {}",
-                         indent,
-                         if let Some(issuer)
-                         = sig.get_issuers().into_iter().next()
-                         {
-                             issuer.to_string()
-                         } else {
-                             "an unknown certificate".into()
-                         })?;
-            }
+        let (reason, message) = sig.reason_for_revocation()
+            .map(|(r, m)| (r, Some(m)))
+            .unwrap_or((ReasonForRevocation::Unspecified, None));
+
+        writeln!(output, "{}                  - {}", indent, reason)?;
+        writeln!(output, "{}                    On: {}",
+                 indent, sig.signature_creation_time()
+                 .expect("valid sigs have one").convert())?;
+        if third_party {
+            writeln!(output, "{}                    Issued by {}",
+                     indent,
+                     if let Some(issuer)
+                     = sig.get_issuers().into_iter().next()
+                     {
+                         issuer.to_string()
+                     } else {
+                         "an unknown certificate".into()
+                     })?;
+        }
+        if let Some(msg) = message {
             writeln!(output, "{}                    Message: {:?}",
                      indent, String::from_utf8_lossy(msg))?;
-        } else {
-            writeln!(output, "{}                  - No reason specified",
-                     indent)?;
-            if third_party {
-                writeln!(output, "{}                    Issued by {}",
-                         indent,
-                         if let Some(issuer)
-                         = &sig.get_issuers().into_iter().next()
-                         {
-                             issuer.to_string()
-                         } else {
-                             "an unknown certificate".into()
-                         })?;
-            }
         }
     }
     Ok(())
@@ -445,10 +518,11 @@ fn inspect_revocation(output: &mut dyn io::Write,
     Ok(())
 }
 
-fn inspect_bare_revocation(output: &mut dyn io::Write, sig: &Signature)
+fn inspect_bare_revocation(sq: &mut Sq,
+                           output: &mut dyn io::Write, sig: &Signature)
                            -> Result<()> {
     let indent = "";
-    inspect_issuers(output, &sig)?;
+    inspect_issuers(sq, output, &sig)?;
     writeln!(output, "{}                 Possible revocation:", indent)?;
     print_reasons(output, indent, false, &[sig])?;
     writeln!(output)?;
@@ -486,7 +560,8 @@ fn inspect_key_flags(flags: openpgp::types::KeyFlags) -> Option<String> {
     }
 }
 
-fn inspect_signatures(output: &mut dyn io::Write,
+fn inspect_signatures(sq: &mut Sq,
+                      output: &mut dyn io::Write,
                       sigs: &[openpgp::packet::Signature]) -> Result<()> {
     use openpgp::types::SignatureType::*;
     for sig in sigs {
@@ -496,7 +571,7 @@ fn inspect_signatures(output: &mut dyn io::Write,
                 writeln!(output, "           Kind: {}", signature_type)?,
         }
 
-        inspect_issuers(output, &sig)?;
+        inspect_issuers(sq, output, &sig)?;
     }
     if ! sigs.is_empty() {
         writeln!(output, "           Note: \
@@ -506,29 +581,34 @@ fn inspect_signatures(output: &mut dyn io::Write,
     Ok(())
 }
 
-fn inspect_issuers(output: &mut dyn io::Write,
+fn inspect_issuers(sq: &mut Sq,
+                   output: &mut dyn io::Write,
                    sig: &Signature) -> Result<()> {
     let mut fps: Vec<_> = sig.issuer_fingerprints().collect();
     fps.sort();
     fps.dedup();
-    let fps: Vec<KeyHandle> = fps.into_iter().map(|fp| fp.into()).collect();
-    for fp in fps.iter() {
-        writeln!(output, " Alleged signer: {}", fp)?;
+    let khs: Vec<KeyHandle> = fps.into_iter().map(|fp| fp.into()).collect();
+    for kh in khs.iter() {
+        writeln!(output, " Alleged signer: {}, {}",
+                 kh, sq.best_userid_for(kh, true))?;
     }
 
     let mut keyids: Vec<_> = sig.issuers().collect();
     keyids.sort();
     keyids.dedup();
     for keyid in keyids {
-        if ! fps.iter().any(|fp| fp.aliases(&keyid.into())) {
-            writeln!(output, " Alleged signer: {}", keyid)?;
+        if ! khs.iter().any(|kh| kh.aliases(&keyid.into())) {
+            writeln!(output, " Alleged signer: {}, {}",
+                     keyid,
+                     sq.best_userid_for(&KeyHandle::from(keyid), true))?;
         }
     }
 
     Ok(())
 }
 
-fn inspect_certifications<'a, A>(output: &mut dyn io::Write,
+fn inspect_certifications<'a, A>(sq: &mut Sq,
+                                 output: &mut dyn io::Write,
                                  policy: &dyn Policy,
                                  certs: A,
                                  print_certifications: bool)
@@ -608,17 +688,21 @@ fn inspect_certifications<'a, A>(output: &mut dyn io::Write,
             let mut fps: Vec<_> = sig.issuer_fingerprints().collect();
             fps.sort();
             fps.dedup();
-            let fps: Vec<KeyHandle> = fps.into_iter().map(|fp| fp.into()).collect();
-            for fp in fps.iter() {
-                writeln!(output, "{}Alleged certifier: {}", indent, fp)?;
+            let khs: Vec<KeyHandle> = fps.into_iter().map(|fp| fp.into()).collect();
+            for kh in khs.iter() {
+                writeln!(output, "{}Alleged certifier: {}, {}",
+                         indent, kh,
+                         sq.best_userid_for(kh, true))?;
             }
             let mut keyids: Vec<_> = sig.issuers().collect();
             keyids.sort();
             keyids.dedup();
             for keyid in keyids {
-                if ! fps.iter().any(|fp| fp.aliases(&keyid.into())) {
-                    writeln!(output, "{}Alleged certifier: {}", indent,
-                             keyid)?;
+                if ! khs.iter().any(|kh| kh.aliases(&keyid.into())) {
+                    writeln!(output, "{}Alleged certifier: {}, {}", indent,
+                             keyid,
+                             sq.best_userid_for(
+                                 &KeyHandle::from(keyid), true))?;
                 }
             }
 
@@ -631,7 +715,8 @@ fn inspect_certifications<'a, A>(output: &mut dyn io::Write,
                          "{}Certification is not valid according to \
                           the current policy:\n\
                           {}  {}",
-                         indent, indent, err)?;
+                         indent, indent,
+                         one_line_error_chain(&err))?;
             }
         }
         if emit_warning {
