@@ -13,7 +13,6 @@ use cert_store::store::StatusUpdate;
 use cert_store::store::StoreError;
 
 use sequoia_wot as wot;
-use wot::store::CertStore;
 use wot::store::Backend;
 use wot::store::Store;
 
@@ -25,14 +24,13 @@ use crate::cli;
 use cli::output::OutputFormat;
 use cli::types::TrustAmount;
 
-use crate::commands::pki as pki_cmd;
-use pki_cmd::output::print_path;
-use pki_cmd::output::print_path_header;
-use pki_cmd::output::print_path_error;
+use output::print_path;
+use output::print_path_header;
+use output::print_path_error;
 #[allow(unused_imports)]
-use pki_cmd::output::OutputType as _;
+use output::OutputType as _;
 
-use crate::Config;
+use crate::Sq;
 
 fn required_trust_amount(trust_amount: Option<TrustAmount<usize>>,
                          certification_network: bool)
@@ -77,8 +75,8 @@ fn have_self_signed_userid(cert: &wot::CertSynopsis,
 }
 
 /// Authenticate bindings defined by a Query on a Network
-fn authenticate(
-    config: &Config,
+fn authenticate<'store, 'rstore>(
+    sq: Sq<'store, 'rstore>,
     precompute: bool,
     list_pattern: Option<String>,
     email: bool,
@@ -87,10 +85,12 @@ fn authenticate(
     trust_amount: Option<TrustAmount<usize>>,
     userid: Option<&UserID>,
     certificate: Option<&KeyHandle>,
+    show_paths: bool,
 ) -> Result<()>
+    where 'store: 'rstore,
 {
     // Build the network.
-    let cert_store = match config.cert_store() {
+    let cert_store = match sq.cert_store() {
         Ok(Some(cert_store)) => cert_store,
         Ok(None) => {
             return Err(anyhow::anyhow!("Certificate store has been disabled"));
@@ -100,37 +100,33 @@ fn authenticate(
         }
     };
 
-    let mut cert_store = CertStore::from_store(
-        cert_store, &config.policy, config.time);
     if precompute {
         cert_store.precompute();
     }
 
-    let n = wot::Network::new(cert_store)?;
-
-    let mut q = wot::QueryBuilder::new(&n);
-    if ! gossip {
-        q.roots(wot::Roots::new(config.trust_roots()));
-    }
+    let mut n = if gossip {
+        wot::NetworkBuilder::rootless(cert_store)
+    } else {
+        wot::NetworkBuilder::rooted(cert_store,
+                                    &*sq.trust_roots())
+    };
     if certification_network {
-        q.certification_network();
+        n = n.certification_network();
     }
-    let q = q.build();
+    let n = n.build();
 
     let required_amount =
         required_trust_amount(trust_amount, certification_network)?;
 
+    let mut certificate_dealiased = None;
     let fingerprint: Option<Fingerprint> = if let Some(kh) = certificate {
-        Some(match kh {
-            KeyHandle::Fingerprint(fpr) => fpr.clone(),
-            kh @ KeyHandle::KeyID(_) => {
-                let certs = q.network().lookup_synopses(kh)?;
-                if certs.is_empty() {
-                    return Err(StoreError::NotFound(kh.clone()).into());
-                }
+        match sq.lookup(std::iter::once(kh), None, true, true) {
+            Ok(certs) => {
+                assert!(certs.len() >= 1);
+
                 if certs.len() > 1 {
                     return Err(anyhow::anyhow!(
-                        "The Key ID {} is ambiguous.  \
+                        "The key ID {} is ambiguous.  \
                          It could refer to any of the following \
                          certificates: {}.",
                         kh,
@@ -140,9 +136,36 @@ fn authenticate(
                             .join(", ")));
                 }
 
-                certs[0].fingerprint()
+                let fpr = certs[0].fingerprint();
+                if ! KeyHandle::from(&fpr).aliases(kh) {
+                    // XXX: If the subkey does not have a backsig,
+                    // then it doesn't authenticate the primary key,
+                    // and thus the authentication confidence must be
+                    // 0, and the certificate should only be shown
+                    // when `--gossip` is passed.
+                    eprintln!("Certificate {} contains the subkey {}.",
+                              fpr, kh);
+                }
+                certificate_dealiased = Some(KeyHandle::from(&fpr));
+                Some(fpr)
             }
-        })
+            Err(err) => {
+                if let Some(StoreError::NotFound(_)) = err.downcast_ref() {
+                    wprintln!("There are no certificates with the \
+                               specified {}.  \
+                               Run `sq network fetch {}` to look for \
+                               matching certificates on public \
+                               directories.",
+                              if let KeyHandle::Fingerprint(_) = kh {
+                                  "fingerprint"
+                              } else {
+                                  "key ID"
+                              },
+                              kh);
+                }
+                return Err(err);
+            }
+        }
     } else {
         None
     };
@@ -173,12 +196,12 @@ fn authenticate(
         // and select the bindings where the User ID matches the email
         // address.
         bindings = if let Some(fingerprint) = fingerprint.as_ref() {
-            q.network().certified_userids_of(fingerprint)
+            n.certified_userids_of(fingerprint)
                 .into_iter()
                 .map(|userid| (fingerprint.clone(), userid))
                 .collect::<Vec<_>>()
         } else {
-            q.network().lookup_synopses_by_email(&email)
+            n.lookup_synopses_by_email(&email)
         };
 
         let email_normalized = userid_check.email_normalized()
@@ -202,7 +225,7 @@ fn authenticate(
             bindings.push((fingerprint, userid.clone()));
         } else {
             // Fingerprint, no User ID.
-            bindings = q.network().certified_userids_of(&fingerprint)
+            bindings = n.certified_userids_of(&fingerprint)
                 .into_iter()
                 .map(|userid| (fingerprint.clone(), userid))
                 .collect();
@@ -210,7 +233,7 @@ fn authenticate(
     } else if let Some(userid) = userid {
         // The caller did not specify a certificate.  Find all
         // bindings with the User ID.
-        bindings = q.network().lookup_synopses_by_userid(userid.clone())
+        bindings = n.lookup_synopses_by_userid(userid.clone())
             .into_iter()
             .map(|fpr| (fpr, userid.clone()))
             .collect();
@@ -218,9 +241,9 @@ fn authenticate(
         // No User ID, no Fingerprint.
         // List everything.
 
-        bindings = q.network().certified_userids();
+        bindings = n.certified_userids();
 
-        if let Some(pattern) = list_pattern {
+        if let Some(ref pattern) = list_pattern {
             // Or rather, just User IDs that match the pattern.
             let pattern = pattern.to_lowercase();
 
@@ -262,24 +285,32 @@ fn authenticate(
     bindings.sort();
     bindings.dedup();
 
-    let mut authenticated = false;
+    let mut authenticated = 0;
     let mut lint_input = true;
 
-    let mut output = match config.output_format {
+    let mut output = match sq.output_format {
         #[cfg(feature = "dot-writer")]
         OutputFormat::DOT => {
             Box::new(output::DotOutputNetwork::new(
                 required_amount,
-                q.roots(),
+                n.roots(),
                 gossip,
                 certification_network,
             ))
             as Box<dyn output::OutputType>
         }
         _ => {
-            Box::new(
-                output::HumanReadableOutputNetwork::new(required_amount, gossip)
-            )
+            if show_paths {
+                Box::new(
+                    output::HumanReadableOutputNetwork::new(
+                        required_amount, gossip))
+                    as Box<dyn output::OutputType>
+            } else {
+                Box::new(
+                    output::ConciseHumanReadableOutputNetwork::new(
+                        &sq, required_amount))
+                    as Box<dyn output::OutputType>
+            }
         }
     };
 
@@ -288,7 +319,7 @@ fn authenticate(
 
         let paths = if gossip {
             // Gossip.
-            let paths = q.gossip(
+            let paths = n.gossip(
                 fingerprint.clone(), userid.clone());
 
             // Sort so the shortest paths come first.
@@ -301,7 +332,7 @@ fn authenticate(
             // This means: exit code is 0, which is what we want when
             // we've found at least one path.
             if paths.len() > 0 {
-                authenticated = true;
+                authenticated += 1;
                 lint_input = false;
             }
 
@@ -309,7 +340,7 @@ fn authenticate(
                 .map(|p| (p, 0))
                 .collect::<Vec<(wot::Path, usize)>>()
         } else {
-            let paths = q.authenticate(
+            let paths = n.authenticate(
                 userid.clone(), fingerprint.clone(), required_amount);
 
             aggregated_amount = paths.amount();
@@ -318,14 +349,13 @@ fn authenticate(
             }
             lint_input = false;
             if aggregated_amount >= required_amount {
-                authenticated = true;
+                authenticated += 1;
             }
 
             paths.into_iter().collect::<Vec<(wot::Path, usize)>>()
         };
 
         output.add_paths(paths, fingerprint, userid, aggregated_amount)?;
-
     }
 
     output.finalize()?;
@@ -333,8 +363,8 @@ fn authenticate(
     // We didn't show anything.  Try to figure out what was wrong.
     if lint_input {
         // See if the target certificate exists.
-        if let Some(kh) = certificate {
-            match q.network().lookup_synopses(kh) {
+        if let Some(kh) = certificate_dealiased {
+            match n.lookup_synopses(&kh) {
                 Err(err) => {
                     wprintln!("Looking up target certificate ({}): {}",
                              kh, err);
@@ -360,7 +390,7 @@ fn authenticate(
 
                         // Check if the certificate has expired.
                         if let Some(e) = cert.expiration_time() {
-                            if e <= q.network().reference_time() {
+                            if e <= n.reference_time() {
                                 wprintln!("Warning: {} is expired.", kh);
                             }
                         }
@@ -376,9 +406,7 @@ fn authenticate(
 
                         // See if there are any certifications made on
                         // this certificate.
-                        if let Ok(cs) = q.network()
-                            .certifications_of(&fpr, 0.into())
-                        {
+                        if let Ok(cs) = n.certifications_of(&fpr, 0.into()) {
                             if cs.iter().all(|cs| {
                                 cs.certifications()
                                     .all(|(_userid, certifications)| {
@@ -417,9 +445,9 @@ fn authenticate(
 
         // See if the trust roots exist.
         if ! gossip {
-            if q.roots().iter().all(|r| {
+            if n.roots().iter().all(|r| {
                 let fpr = r.fingerprint();
-                if let Err(err) = q.network().lookup_synopsis_by_fpr(&fpr) {
+                if let Err(err) = n.lookup_synopsis_by_fpr(&fpr) {
                     wprintln!("Looking up trust root ({}): {}.",
                              fpr, err);
                     true
@@ -433,20 +461,74 @@ fn authenticate(
         }
     }
 
-    if ! authenticated {
-        if ! lint_input {
-            wprintln!("Could not authenticate any paths.");
+    let pattern = || {
+        certificate.map(|kh| kh.to_string())
+            .or_else(|| userid.map(|u| {
+                String::from_utf8_lossy(u.value()).to_string()
+            }))
+            .or_else(|| list_pattern.clone())
+    };
+
+    if gossip {
+        // We are in gossip mode.  Mention `sq pki link` as a way to
+        // mark bindings as authenticated.
+        if ! bindings.is_empty() {
+            wprintln!("After checking that a user ID really belongs to \
+                       a certificate, use `sq pki link add` to mark \
+                       the binding as authenticated, or use \
+                       `sq network fetch FINGERPRINT|EMAIL` to look for \
+                       new certifications.");
+        }
+    } else if bindings.is_empty() {
+        // There are no matching bindings.  Tell the user about `sq
+        // network fetch`.
+        if let Some(pattern) = pattern() {
+            wprintln!("No bindings match.  Try using \
+                       `sq network fetch {:?}` to search public directories \
+                       for matching certificates.",
+                      pattern);
         } else {
-            wprintln!("No paths found.");
+            wprintln!("The certificate store does not contain any \
+                       certificates.");
+        }
+    } else if bindings.len() - authenticated > 0 {
+        // Some of the matching bindings are unauthenticated.  Tell
+        // the user about the `--gossip` option.
+        let bindings = bindings.len();
+        assert!(bindings > 0);
+        let unauthenticated = bindings - authenticated;
+
+        if bindings == 1 {
+            wprintln!("1 binding found.");
+        } else {
+            wprintln!("{} bindings found.", bindings);
+        }
+
+        if unauthenticated == 1 {
+            wprintln!("Skipped 1 binding, which could not be authenticated.");
+            wprintln!("Pass `--gossip` to see the unauthenticated binding.");
+        } else {
+            wprintln!("Skipped {} bindings, which could not be authenticated.",
+                      unauthenticated);
+            wprintln!("Pass `--gossip` to see the unauthenticated bindings.");
         }
-        std::process::exit(1);
     }
 
-    Ok(())
+    if authenticated == 0 {
+        if ! lint_input {
+            Err(anyhow::anyhow!("Could not authenticate any paths."))
+        } else if bindings.is_empty() {
+            Err(anyhow::anyhow!("No bindings match the query."))
+        } else {
+            Err(anyhow::anyhow!("Could not authenticate any matching bindings."))
+        }
+    } else {
+        Ok(())
+    }
 }
 
 // For `sq-wot path`.
-fn check_path(config: &Config,
+fn check_path(sq: &Sq,
               gossip: bool,
               certification_network: bool,
               trust_amount: Option<TrustAmount<usize>>,
@@ -456,7 +538,7 @@ fn check_path(config: &Config,
     tracer!(TRACE, "check_path");
 
     // Build the network.
-    let cert_store = match config.cert_store() {
+    let cert_store = match sq.cert_store() {
         Ok(Some(cert_store)) => cert_store,
         Ok(None) => {
             return Err(anyhow::anyhow!("Certificate store has been disabled"));
@@ -466,19 +548,16 @@ fn check_path(config: &Config,
         }
     };
 
-    let cert_store = CertStore::from_store(
-        cert_store, &config.policy, config.time);
-
-    let n = wot::Network::new(cert_store)?;
-
-    let mut q = wot::QueryBuilder::new(&n);
-    if ! gossip {
-        q.roots(wot::Roots::new(config.trust_roots()));
-    }
+    let mut n = if gossip {
+        wot::NetworkBuilder::rootless(cert_store)
+    } else {
+        wot::NetworkBuilder::rooted(cert_store,
+                                    &*sq.trust_roots())
+    };
     if certification_network {
-        q.certification_network();
+        n = n.certification_network();
     }
-    let q = q.build();
+    let q = n.build();
 
     let required_amount =
         required_trust_amount(trust_amount, certification_network)?;
@@ -486,13 +565,13 @@ fn check_path(config: &Config,
     let (khs, userid) = (path.certs()?, path.userid()?);
     assert!(khs.len() > 0, "guaranteed by clap");
 
-    let r = q.lint_path(&khs, &userid, required_amount, &config.policy);
+    let r = q.lint_path(&khs, &userid, required_amount, sq.policy);
 
     let target_kh = khs.last().expect("have one");
 
     match r {
         Ok(path) => {
-            match config.output_format {
+            match sq.output_format {
                 #[cfg(feature = "dot-writer")]
                 OutputFormat::DOT => {
                     wprintln!(
@@ -506,7 +585,7 @@ fn check_path(config: &Config,
                         path.amount(),
                         required_amount,
                     );
-                    print_path(&path, &userid, "  ");
+                    print_path(&path, &userid, "  ")?;
                 }
             };
 
@@ -515,7 +594,7 @@ fn check_path(config: &Config,
             }
         }
         Err(err) => {
-            match config.output_format {
+            match sq.output_format {
                 #[cfg(feature = "dot-writer")]
                 OutputFormat::DOT => {
                     wprintln!(
@@ -547,7 +626,7 @@ impl StatusListener for KeyServerUpdate {
     }
 }
 
-pub fn dispatch(config: Config, cli: cli::pki::Command) -> Result<()> {
+pub fn dispatch(sq: Sq, cli: cli::pki::Command) -> Result<()> {
     tracer!(TRACE, "pki::dispatch");
 
     use cli::pki::*;
@@ -555,54 +634,67 @@ pub fn dispatch(config: Config, cli: cli::pki::Command) -> Result<()> {
         // Authenticate a given binding.
         Subcommands::Authenticate(AuthenticateCommand {
             email, gossip, certification_network, trust_amount,
-            cert, userid,
+            cert, userid, show_paths,
         }) => authenticate(
-            &config, false, None,
+            sq, false, None,
             *email, *gossip, *certification_network, *trust_amount,
-            Some(&userid), Some(&cert))?,
+            Some(&userid), Some(&cert), *show_paths,
+        )?,
 
         // Find all authenticated bindings for a given User ID, list
         // the certificates.
         Subcommands::Lookup(LookupCommand {
             email, gossip, certification_network, trust_amount,
-            userid,
+            userid, show_paths,
         }) => authenticate(
-            &config, false, None,
+            sq, false, None,
             *email, *gossip, *certification_network, *trust_amount,
-            Some(&userid), None)?,
+            Some(&userid), None, *show_paths)?,
 
         // Find and list all authenticated bindings for a given
         // certificate.
         Subcommands::Identify(IdentifyCommand {
             gossip, certification_network, trust_amount,
-            cert,
+            cert, show_paths,
         }) => authenticate(
-            &config, false, None,
+            sq, false, None,
             false, *gossip, *certification_network, *trust_amount,
-            None, Some(&cert))?,
+            None, Some(&cert), *show_paths)?,
 
         // List all authenticated bindings.
         Subcommands::List(ListCommand {
             email, gossip, certification_network, trust_amount,
-            pattern,
-        }) => authenticate(
-            &config, pattern.is_none(), pattern,
-            *email, *gossip, *certification_network, *trust_amount,
-            None, None)?,
+            pattern, show_paths,
+        }) => if let Some(handle) = pattern.as_ref()
+            .and_then(|p| p.parse().ok())
+            .iter().filter(|_| ! *email).next()
+        {
+            // A key handle was given as pattern and --email was not
+            // given.  Act like `sq pki identify`.
+            authenticate(
+                sq, false, None,
+                false, *gossip, *certification_network, *trust_amount,
+                None, Some(&handle), *show_paths)?;
+        } else {
+            authenticate(
+                sq, pattern.is_none(), pattern,
+                *email, *gossip, *certification_network, *trust_amount,
+                None, None, *show_paths)?;
+        },
 
         // Authenticates a given path.
         Subcommands::Path(PathCommand {
             gossip, certification_network, trust_amount,
             path,
         }) => check_path(
-            &config, *gossip, *certification_network, *trust_amount,
+            &sq, *gossip, *certification_network, *trust_amount,
             path)?,
 
         Subcommands::Certify(command) =>
-            self::certify::certify(config, command)?,
+            self::certify::certify(sq, command)?,
 
         Subcommands::Link(command) =>
-            self::link::link(config, command)?,
+            self::link::link(sq, command)?,
     }
 
     Ok(())
