@@ -2,11 +2,11 @@ use anyhow::Context as _;
 use std::fs;
 use std::io;
 use std::path::PathBuf;
-use std::time::SystemTime;
 use tempfile::NamedTempFile;
 
 use sequoia_openpgp as openpgp;
 use openpgp::armor;
+use openpgp::KeyHandle;
 use openpgp::{Packet, Result};
 use openpgp::packet::prelude::*;
 use openpgp::packet::signature::subpacket::NotationData;
@@ -18,9 +18,11 @@ use openpgp::serialize::Serialize;
 use openpgp::serialize::stream::{
     Message, Armorer, Signer, LiteralWriter,
 };
+use openpgp::types::KeyFlags;
 use openpgp::types::SignatureType;
 
-use crate::Config;
+
+use crate::Sq;
 use crate::load_certs;
 use crate::parse_notations;
 
@@ -31,7 +33,7 @@ use crate::cli::types::FileOrStdout;
 mod merge_signatures;
 use merge_signatures::merge_signatures;
 
-pub fn dispatch(config: Config, command: cli::sign::Command) -> Result<()> {
+pub fn dispatch(sq: Sq, command: cli::sign::Command) -> Result<()> {
     tracer!(TRACE, "sign::dispatch");
 
     let mut input = command.input.open()?;
@@ -44,16 +46,15 @@ pub fn dispatch(config: Config, command: cli::sign::Command) -> Result<()> {
         return Err(anyhow::anyhow!("Notarizing messages is not supported."));
     }
 
-    let private_key_store = command.private_key_store.as_deref();
     let secrets =
         load_certs(command.secret_key_file.iter().map(|s| s.as_ref()))?;
-    let time = Some(config.time);
+    let signer_keys = &command.signer_key[..];
 
     let notations = parse_notations(command.notation)?;
 
     if let Some(merge) = command.merge {
         let output = output.create_pgp_safe(
-            config.force,
+            sq.force,
             binary,
             armor::Kind::Message,
         )?;
@@ -61,75 +62,68 @@ pub fn dispatch(config: Config, command: cli::sign::Command) -> Result<()> {
         let mut input2 = data.open()?;
         merge_signatures(&mut input, &mut input2, output)?;
     } else if command.clearsign {
-        let output = output.create_safe(config.force)?;
-        clearsign(config, private_key_store, input, output, secrets,
-                  time, &notations)?;
+        let output = output.create_safe(sq.force)?;
+        clearsign(sq, input, output, secrets, &notations)?;
     } else {
-        sign(config,
-             private_key_store,
+        sign(sq,
              &mut input,
              output,
              secrets,
+             signer_keys,
              detached,
              binary,
              append,
              notarize,
-             time,
              &notations)?;
     }
 
     Ok(())
 }
 
-pub fn sign<'a, 'certdb>(
-    config: Config<'certdb>,
-    private_key_store: Option<&'a str>,
+pub fn sign<'a, 'store, 'rstore>(
+    sq: Sq<'store, 'rstore>,
     input: &'a mut (dyn io::Read + Sync + Send),
     output: &'a FileOrStdout,
     secrets: Vec<openpgp::Cert>,
+    signer_keys: &[KeyHandle],
     detached: bool,
     binary: bool,
     append: bool,
     notarize: bool,
-    time: Option<SystemTime>,
     notations: &'a [(bool, NotationData)])
     -> Result<()>
 {
     match (detached, append|notarize) {
         (_, false) | (true, true) =>
-            sign_data(config,
-                      private_key_store,
+            sign_data(sq,
                       input,
                       output,
                       secrets,
+                      signer_keys,
                       detached,
                       binary,
                       append,
-                      time,
                       notations),
         (false, true) =>
-            sign_message(config,
-                         private_key_store,
+            sign_message(sq,
                          input,
                          output,
                          secrets,
                          binary,
                          notarize,
-                         time,
                          notations),
     }
 }
 
-fn sign_data<'a, 'certdb>(
-    config: Config<'certdb>,
-    private_key_store: Option<&'a str>,
+fn sign_data<'a, 'store, 'rstore>(
+    sq: Sq<'store, 'rstore>,
     input: &'a mut (dyn io::Read + Sync + Send),
     output_path: &'a FileOrStdout,
-    secrets: Vec<openpgp::Cert>,
+    mut secrets: Vec<openpgp::Cert>,
+    signer_keys: &[KeyHandle],
     detached: bool,
     binary: bool,
     append: bool,
-    time: Option<SystemTime>,
     notations: &'a [(bool, NotationData)])
     -> Result<()>
 {
@@ -164,12 +158,17 @@ fn sign_data<'a, 'certdb>(
             let tmp_path = tmp_file.path().into();
             (Box::new(tmp_file), sigs, Some(tmp_path))
         } else {
-            (output_path.create_safe(config.force)?, Vec::new(), None)
+            (output_path.create_safe(sq.force)?, Vec::new(), None)
         };
 
-    let mut keypairs = super::get_signing_keys(
-        &secrets, &config.policy, private_key_store, time, None)?;
-    if keypairs.is_empty() {
+    for kh in signer_keys.into_iter() {
+        let cert = sq.lookup_one(
+            kh, Some(KeyFlags::empty().set_signing()), true)?;
+        secrets.push(cert);
+    };
+
+    let mut signers = sq.get_signing_keys(&secrets, None)?;
+    if signers.is_empty() {
         return Err(anyhow::anyhow!("No signing keys found"));
     }
 
@@ -203,11 +202,11 @@ fn sign_data<'a, 'certdb>(
     }
 
     let mut signer = Signer::with_template(
-        message, keypairs.pop().unwrap().0, builder);
-    if let Some(time) = time {
-        signer = signer.creation_time(time);
-    }
-    for s in keypairs {
+        message,
+        signers.pop().unwrap().0,
+        builder);
+    signer = signer.creation_time(sq.time);
+    for s in signers {
         signer = signer.add_signer(s.0);
     }
     if detached {
@@ -242,48 +241,41 @@ fn sign_data<'a, 'certdb>(
     Ok(())
 }
 
-fn sign_message<'a, 'certdb>(
-    config: Config<'certdb>,
-    private_key_store: Option<&'a str>,
+fn sign_message<'a, 'store, 'rstore>(
+    sq: Sq<'store, 'rstore>,
     input: &'a mut (dyn io::Read + Sync + Send),
     output: &'a FileOrStdout,
     secrets: Vec<openpgp::Cert>,
     binary: bool,
     notarize: bool,
-    time: Option<SystemTime>,
     notations: &'a [(bool, NotationData)])
     -> Result<()>
 {
     let mut output = output.create_pgp_safe(
-        config.force,
+        sq.force,
         binary,
         armor::Kind::Message,
     )?;
-    sign_message_(config,
-                  private_key_store,
+    sign_message_(sq,
                   input,
                   secrets,
                   notarize,
-                  time,
                   notations,
                   &mut output)?;
     output.finalize()?;
     Ok(())
 }
 
-fn sign_message_<'a, 'certdb>(
-    config: Config<'certdb>,
-    private_key_store: Option<&'a str>,
+fn sign_message_<'a, 'store, 'rstore>(
+    sq: Sq<'store, 'rstore>,
     input: &'a mut (dyn io::Read + Sync + Send),
     secrets: Vec<openpgp::Cert>,
     notarize: bool,
-    time: Option<SystemTime>,
     notations: &'a [(bool, NotationData)],
     output: &mut (dyn io::Write + Sync + Send))
     -> Result<()>
 {
-    let mut keypairs = super::get_signing_keys(
-        &secrets, &config.policy, private_key_store, time, None)?;
+    let mut keypairs = sq.get_signing_keys(&secrets, None)?;
     if keypairs.is_empty() {
         return Err(anyhow::anyhow!("No signing keys found"));
     }
@@ -363,9 +355,7 @@ fn sign_message_<'a, 'certdb>(
 
                 let mut signer = Signer::with_template(
                     sink, keypairs.pop().unwrap().0, builder);
-                if let Some(time) = time {
-                    signer = signer.creation_time(time);
-                }
+                signer = signer.creation_time(sq.time);
                 for s in keypairs.drain(..) {
                     signer = signer.add_signer(s.0);
                 }
@@ -480,17 +470,14 @@ fn sign_message_<'a, 'certdb>(
     Ok(())
 }
 
-pub fn clearsign(config: Config,
-                 private_key_store: Option<&str>,
+pub fn clearsign(sq: Sq,
                  mut input: impl io::Read + Sync + Send,
                  mut output: impl io::Write + Sync + Send,
                  secrets: Vec<openpgp::Cert>,
-                 time: Option<SystemTime>,
                  notations: &[(bool, NotationData)])
                  -> Result<()>
 {
-    let mut keypairs = super::get_signing_keys(
-        &secrets, &config.policy, private_key_store, time, None)?;
+    let mut keypairs = sq.get_signing_keys(&secrets, None)?;
     if keypairs.is_empty() {
         return Err(anyhow::anyhow!("No signing keys found"));
     }
@@ -509,9 +496,7 @@ pub fn clearsign(config: Config,
     let mut signer = Signer::with_template(
         message, keypairs.pop().unwrap().0, builder)
         .cleartext();
-    if let Some(time) = time {
-        signer = signer.creation_time(time);
-    }
+    signer = signer.creation_time(sq.time);
     for s in keypairs {
         signer = signer.add_signer(s.0);
     }
