@@ -1,16 +1,23 @@
 use std::io::{self, Read};
 
 use sequoia_openpgp as openpgp;
-use self::openpgp::types::SymmetricAlgorithm;
+use openpgp::armor::ReaderMode;
 use self::openpgp::fmt::hex;
 use self::openpgp::crypto::mpi;
-use self::openpgp::{Packet, Result};
+use self::openpgp::{Cert, Packet, Result};
 use self::openpgp::packet::prelude::*;
 use self::openpgp::packet::header::CTB;
 use self::openpgp::packet::{Header, header::BodyLength, Signature};
 use self::openpgp::packet::signature::subpacket::{Subpacket, SubpacketValue};
 use self::openpgp::crypto::S2K;
-use self::openpgp::parse::{map::Map, Parse, PacketParserResult};
+use self::openpgp::parse::{
+    Dearmor,
+    Parse,
+    PacketParserBuilder,
+    PacketParserResult,
+    map::Map,
+    stream::DecryptionHelper,
+};
 
 use crate::Convert;
 use crate::cli::types::SessionKey;
@@ -26,23 +33,51 @@ pub enum Kind {
 }
 
 #[allow(clippy::redundant_pattern_matching)]
-pub fn dump<W>(input: &mut (dyn io::Read + Sync + Send),
+pub fn dump<W>(sq: &crate::Sq,
+               secrets: Vec<Cert>,
+               input: &mut (dyn io::Read + Sync + Send),
                output: &mut dyn io::Write,
                mpis: bool, hex: bool,
-               sk: Option<&SessionKey>,
+               session_keys: Vec<SessionKey>,
                width: W)
                -> Result<Kind>
     where W: Into<Option<usize>>
 {
     let mut ppr
-        = self::openpgp::parse::PacketParserBuilder::from_reader(input)?
-        .map(hex).build()?;
-    let mut message_encrypted = false;
+        = self::openpgp::parse::PacketParserBuilder::from_reader(input)?;
+
+    // To produce hex dumps, we need to enable mapping, but also turn
+    // on buffering.  This makes sure that the map contains the whole
+    // packet content, even if it has not been parsed (such as when
+    // encountering unknown or junk pseudo packets).
+    if hex {
+        ppr = ppr.map(true).buffer_unread_content();
+    }
+
+    let mut ppr = ppr.build()?;
+
     let width = width.into().unwrap_or(80);
+    let mut first_armor_block = true;
+    let mut is_keyring = true;
+    let mut helper = crate::commands::decrypt::Helper::new(
+        &sq, 0, Vec::new(), secrets, session_keys.clone(), false);
+
+  loop {
     let mut dumper = PacketDumper::new(width, mpis);
+    let mut message_encrypted = false;
+    let mut pkesks = vec![];
+    let mut skesks = vec![];
 
     while let PacketParserResult::Some(mut pp) = ppr {
         let additional_fields = match pp.packet {
+            Packet::PKESK(ref p) => {
+                pkesks.push(p.clone());
+                vec![]
+            },
+            Packet::SKESK(ref p) => {
+                skesks.push(p.clone());
+                vec![]
+            },
             Packet::Literal(_) => {
                 let mut prefix = vec![0; 40];
                 let n = pp.read(&mut prefix)?;
@@ -52,61 +87,44 @@ pub fn dump<W>(input: &mut (dyn io::Read + Sync + Send),
                             if n == prefix.len() { "..." } else { "" }),
                 ]
             },
-            Packet::SEIP(_) if sk.is_none() => {
+            Packet::SEIP(_) | Packet::AED(_) => {
+                dumper.flush(output)?;
                 message_encrypted = true;
-                vec!["No session key supplied".into()]
-            }
-            Packet::SEIP(_) if sk.is_some() => {
-                message_encrypted = true;
-                let sk = sk.unwrap();
-                let decrypted_with = if let Some(algo) = sk.symmetric_algo {
-                    // We know which algorithm to use, so only try decrypting
-                    // with that one.
-                    pp.decrypt(algo, &sk.session_key).is_ok().then(|| algo)
+
+                let mut success = false;
+                let mut fields = Vec::new();
+                let sym_algo_hint = if let Packet::AED(ref aed) = pp.packet {
+                    Some(aed.symmetric_algo())
                 } else {
-                    // We don't know which algorithm to use,
-                    // try to find one that decrypts the message.
-                    (1u8..=19)
-                        .map(SymmetricAlgorithm::from)
-                        .find(|algo| pp.decrypt(*algo, &sk.session_key).is_ok())
+                    None
+                };
+                let decryption_proxy = |algo, secret: &_| {
+                    // Take the algo from the AED packet over
+                    // the dummy one from the SKESK5 packet.
+                    let algo = sym_algo_hint.unwrap_or(algo);
+                    let result = pp.decrypt(algo, secret);
+                    if let Ok(_) = result {
+                        fields.push(
+                            format!("Session key: {}", &hex::encode(secret)));
+                        fields.push(format!("Symmetric algo: {}", algo));
+                        fields.push("Decryption successful".into());
+                        success = true;
+                        true
+                    } else {
+                        false
+                    }
                 };
 
-                let mut fields = Vec::new();
-                fields.push(format!("Session key: {}", &sk.display_sensitive()));
-                if let Some(algo) = decrypted_with {
-                    fields.push(format!("Symmetric algo: {}", algo));
-                    fields.push("Decryption successful".into());
-                } else {
-                    if let Some(algo) = sk.symmetric_algo {
+                if let Err(e) = helper.decrypt(&pkesks[..], &skesks[..],
+                                               sym_algo_hint, decryption_proxy)
+                {
+                    for algo in session_keys.iter()
+                        .filter_map(|sk| sk.symmetric_algo)
+                    {
                         fields.push(format!(
-                            "Indicated Symmetric algo: {}", algo
-                        ));
-                    };
-                    fields.push("Decryption failed".into());
-                }
-                fields
-            },
-            Packet::AED(_) if sk.is_none() => {
-                message_encrypted = true;
-                vec!["No session key supplied".into()]
-            }
-            Packet::AED(_) if sk.is_some() => {
-                message_encrypted = true;
-                let sk = sk.as_ref().unwrap();
-                let algo = if let Packet::AED(ref aed) = pp.packet {
-                    aed.symmetric_algo()
-                } else {
-                    unreachable!()
-                };
-
-                let _ = pp.decrypt(algo, &sk.session_key);
-
-                let mut fields = Vec::new();
-                fields.push(format!("Session key: {}", sk.display_sensitive()));
-                if pp.processed() {
-                    fields.push("Decryption successful".into());
-                } else {
-                    fields.push("Decryption failed".into());
+                            "Indicated Symmetric algo: {}", algo));
+                    }
+                    fields.push(format!("Decryption failed: {}", e));
                 }
                 fields
             },
@@ -135,20 +153,53 @@ pub fn dump<W>(input: &mut (dyn io::Read + Sync + Send),
     dumper.flush(output)?;
 
     if let PacketParserResult::EOF(eof) = ppr {
-        if eof.is_message().is_ok() {
-            Ok(Kind::Message {
-                encrypted: message_encrypted,
-            })
-        } else if eof.is_cert().is_ok() {
-            Ok(Kind::Cert)
-        } else if eof.is_keyring().is_ok() {
-            Ok(Kind::Keyring)
-        } else {
-            Ok(Kind::Unknown)
+        let is_message = eof.is_message().is_ok() && first_armor_block;
+        let is_cert = eof.is_cert().is_ok() && first_armor_block;
+        is_keyring &= eof.is_keyring().is_ok();
+        first_armor_block = false;
+
+        // Now, the parser is exhausted, but we may find another
+        // armored blob.  Note that this can only happen if the first
+        // set of packets was also armored.
+        match PacketParserBuilder::from_buffered_reader(eof.into_reader())
+            .and_then(
+                |builder| builder
+                    .dearmor(Dearmor::Enabled(
+                        ReaderMode::Tolerant(None)))
+                    .build())
+        {
+            Ok(ppr_) => {
+                writeln!(output, "Note: There is another block of armored \
+                                  OpenPGP data.")?;
+
+                if is_message {
+                    writeln!(output, "Note: Data concatenated to a message is \
+                                      likely an error.")?;
+                } else if is_cert || is_keyring {
+                    writeln!(output, "Note: This is a non-standard extension \
+                                      to OpenPGP.")?;
+                }
+                writeln!(output)?;
+
+                ppr = ppr_;
+                continue;
+            },
+            Err(_) => break if is_message {
+                Ok(Kind::Message {
+                    encrypted: message_encrypted,
+                })
+            } else if is_cert {
+                Ok(Kind::Cert)
+            } else if is_keyring {
+                Ok(Kind::Keyring)
+            } else {
+                Ok(Kind::Unknown)
+            }
         }
     } else {
         unreachable!()
     }
+  }
 }
 
 struct Node {
@@ -577,7 +628,7 @@ impl PacketDumper {
             Literal(ref l) => {
                 writeln!(output, "{}  Format: {}", i, l.format())?;
                 if let Some(filename) = l.filename() {
-                    writeln!(output, "{}  Filename: {}", i,
+                    writeln!(output, "{}  Filename: {:?}", i,
                              String::from_utf8_lossy(filename))?;
                 }
                 if let Some(timestamp) = l.date() {
@@ -702,11 +753,16 @@ impl PacketDumper {
         writeln!(output, "{}", i)?;
 
         if let Some(map) = map {
+            if map.iter().next().is_none() {
+                // There is no data, we cannot dump anything.
+                return Ok(());
+            }
+
             let mut hd = hex::Dumper::new(output, self.indentation_for_hexdump(
                 i, map.iter()
                     .map(|f| if f.name() == "body" { 16 } else { f.name().len() })
                     .max()
-                    .expect("we always have one entry")));
+                    .expect("we checked that there is one entry")));
 
             for field in map.iter() {
                 if field.name() == "body" {
