@@ -26,13 +26,15 @@
 use roff::{bold, italic, roman, Inline, Roff};
 use std::path::{Path, PathBuf};
 
+use clap::builder::ValueRange;
+
 /// The "manual" the manual page is meant for. The full Unix
 /// documentation is (or was) divided into separate manuals, some of
 /// which don't consist of manual pages.
 const MANUAL: &str = "User Commands";
 
 /// The "source" of the manual: who produced the manual.
-const SOURCE: &str = "Sequoia-PGP";
+const SOURCE: &str = "Sequoia PGP";
 
 /// Text to add to the end of the "SEE ALSO" section of sq manual page.
 const SEE_ALSO: &str = "For the full documentation see <https://book.sequoia-pgp.org>.";
@@ -396,6 +398,14 @@ impl Command {
         man.section("DESCRIPTION");
         man.text_with_period(&self.description());
 
+        if self == &builder.maincmd && self.has_options() {
+            man.section("OPTIONS");
+            man.subsection("Global options");
+            for opt in self.get_options().iter() {
+                man.option(opt);
+            }
+        }
+
         man.section("SUBCOMMANDS");
         for sub in &self.subcommands {
             let desc = sub.description();
@@ -453,10 +463,8 @@ impl Command {
             .filter(|o| ! builder.maincmd.get_options().contains(o))
             .next().is_some();
 
-        if main_opts || self_opts {
-            man.section("OPTIONS");
-        }
         if self_opts {
+            man.section("OPTIONS");
             man.subsection("Subcommand options");
             for opt in self.get_options().iter()
                 .filter(|o| ! builder.maincmd.get_options().contains(o))
@@ -466,9 +474,11 @@ impl Command {
         }
         if main_opts {
             man.subsection("Global options");
-            for opt in builder.maincmd.get_options().iter() {
-                man.option(opt);
-            }
+            man.roff.text(vec![
+                roman("See "),
+                bold("sq"), roman("("), roman("1"), roman(")"),
+                roman(" for a description of the global options."),
+            ]);
         }
 
         man.exit_status_section(self);
@@ -499,10 +509,29 @@ struct CommandOption {
     index: Option<usize>,
     short: Option<String>,
     long: Option<String>,
-    value_names: Option<Vec<String>>,
+    value_names: Values,
+    possible_values: Vec<String>,
+    default_values: Vec<String>,
     help: Option<String>,
 }
 
+/// Formal arguments for a command option.
+#[derive(Clone, Debug, PartialEq, Eq)]
+enum Values {
+    /// This option does not take any arguments.
+    None,
+
+    /// This option takes the given arguments.
+    Some(Vec<String>),
+
+    /// This option optionally takes the given arguments.
+    ///
+    /// XXX: Currently, this is all or nothing, i.e. either all
+    /// arguments are mandatory, or all are optional.  Let's see
+    /// whether we need something more complicated.
+    Optional(Vec<String>),
+}
+
 impl CommandOption {
     /// Return a key for sorting a list of options.
     ///
@@ -529,14 +558,23 @@ impl CommandOption {
     /// Create a `CommandOption` from a `clap::Arg`.
     fn from_arg(arg: &clap::Arg) -> Self {
         let value_names = if arg.get_num_args()
-            .map(|r| r == clap::builder::ValueRange::EMPTY)
+            .map(|r| r == ValueRange::EMPTY)
             .unwrap_or(true)
         {
-            None
+            Values::None
         } else if let Some(names) = arg.get_value_names() {
-            Some(names.iter().map(|name| name.to_string()).collect())
+            let names =
+                names.iter().map(|name| name.to_string()).collect();
+
+            if arg.get_num_args()
+                .map(|r| r.min_values() == 0).unwrap_or(false)
+            {
+                Values::Optional(names)
+            } else {
+                Values::Some(names)
+            }
         } else {
-            None
+            Values::None
         };
 
         Self {
@@ -544,6 +582,19 @@ impl CommandOption {
             short: arg.get_short().map(|o| format!("-{}", o)),
             long: arg.get_long().map(|o| format!("--{}", o)),
             value_names,
+            possible_values: arg.get_possible_values().iter()
+                .map(|o| o.get_name().into()).collect(),
+            default_values: if arg.get_num_args()
+                .map(|r| r.max_values() > 0).unwrap_or(false) {
+                    // At least one argument.
+                    arg.get_default_values().iter()
+                        .map(|s| s.to_string_lossy().to_string())
+                        .collect()
+                } else {
+                    // No argument, i.e. a flag.  We don't want to
+                    // output `[default: false]` in that case.
+                    vec![]
+                },
             help: arg.get_long_help().or(arg.get_help())
                 .map(|s| s.to_string()),
         }
@@ -659,21 +710,60 @@ impl ManualPage {
             line.push(bold(long));
         }
 
-        if let Some(values) = &opt.value_names {
-            if (opt.short.is_some() || opt.long.is_some())
-                && values.len() == 1
-            {
-                line.push(roman("="));
-                line.push(italic(&values[0]));
-            } else {
-                for value in values {
-                    line.push(roman(" "));
-                    line.push(italic(value));
+        match &opt.value_names {
+            Values::None => (),
+            Values::Some(values) | Values::Optional(values) => {
+                if matches!(opt.value_names, Values::Optional(_)) {
+                    line.push(roman("["));
                 }
-            }
+
+                if (opt.short.is_some() || opt.long.is_some())
+                    && values.len() == 1
+                {
+                    line.push(roman("="));
+                    line.push(italic(&values[0]));
+                } else {
+                    for value in values {
+                        line.push(roman(" "));
+                        line.push(italic(value));
+                    }
+                }
+
+                if matches!(opt.value_names, Values::Optional(_)) {
+                    line.push(roman("]"));
+                }
+            },
         }
 
         self.tagged_paragraph(line, &opt.help);
+
+        if ! opt.default_values.is_empty() {
+            self.indented_paragraph();
+            let mut line = vec![];
+            line.push(roman("[default: "));
+            for (i, v) in opt.default_values.iter().enumerate() {
+                if i > 0 {
+                    line.push(roman(", "));
+                }
+                line.push(bold(v));
+            }
+            line.push(roman("]"));
+            self.roff.text(line);
+        }
+
+        if ! opt.possible_values.is_empty() {
+            self.indented_paragraph();
+            let mut line = vec![];
+            line.push(roman("[possible values: "));
+            for (i, v) in opt.possible_values.iter().enumerate() {
+                if i > 0 {
+                    line.push(roman(", "));
+                }
+                line.push(bold(v));
+            }
+            line.push(roman("]"));
+            self.roff.text(line);
+        }
     }
 
     /// Typeset an EXIT STATUS section, if available.
@@ -739,7 +829,7 @@ impl ManualPage {
                             if line.len() > EXAMPLE_COMMAND_MAX_WIDTH {
                                 warn!("Command in example exceeds {} chars:",
                                       EXAMPLE_COMMAND_MAX_WIDTH);
-                                fail!("{}", line);
+                                fail!("{} ({} chars)", line, line.len());
                             }
                             self.code(true);
                             self.roff.control("RS", []);
@@ -750,7 +840,7 @@ impl ManualPage {
                             if line.len() > EXAMPLE_CONTINUATION_MAX_WIDTH {
                                 warn!("Continuation in example exceeds {} chars:",
                                       EXAMPLE_CONTINUATION_MAX_WIDTH);
-                                fail!("{}", line);
+                                fail!("{} ({} chars)", line, line.len());
                             }
                             self.code(true);
                             self.roff.control("RS", []);
@@ -805,6 +895,11 @@ impl ManualPage {
         self.roff.control("PP", []);
     }
 
+    /// Start a new indented paragraph with the IP troff command.
+    fn indented_paragraph(&mut self) {
+        self.roff.control("IP", []);
+    }
+
     /// Start a tagged paragraph with th TP troff command.
     ///
     /// This command takes the line after the command and typesets it,
