@@ -27,9 +27,11 @@ mod integration {
         // Build up the command line.
         let mut cmd = Command::cargo_bin("sq")?;
         cmd.args(["--no-cert-store",
+                  "--no-key-store",
                   "key", "generate",
+                  "--no-userids",
                   "--time", iso8601,
-                  "--expiry", "never",
+                  "--expiration", "never",
                   "--output", &*key_pgp.to_string_lossy()]);
 
         cmd.assert().success();
