@@ -60,9 +60,11 @@ fn sq_gen_key(cert_store: Option<&str>, userids: &[&str], file: &str) -> Cert
 {
     let mut cmd = Command::cargo_bin("sq").expect("have sq");
     cmd.args(["--no-cert-store",
+              "--no-key-store",
               "key", "generate",
+              "--allow-non-canonical-userids",
               "--time", &tick(),
-              "--expiry", "never",
+              "--expiration", "never",
               "--output", file]);
     for userid in userids.iter() {
         cmd.args(["--userid", userid]);
@@ -86,6 +88,7 @@ fn sq_verify(cert_store: Option<&str>,
              good_sigs: usize, good_checksums: usize)
 {
     let mut cmd = Command::cargo_bin("sq").expect("have sq");
+    cmd.arg("--no-key-store");
     if let Some(cert_store) = cert_store {
         cmd.args(&["--cert-store", cert_store]);
     } else {
@@ -192,7 +195,8 @@ fn sq_certify(cert_store: &str,
 {
     let mut cmd = Command::cargo_bin("sq").expect("have sq");
     cmd.args(&["--cert-store", cert_store]);
-    cmd.args(&["pki", "certify", "--time", &tick(), key, cert, userid]);
+    cmd.args(&["pki", "certify", "--time", &tick(),
+               "--certifier-file", key, cert, userid]);
     if let Some(trust_amount) = trust_amount {
         cmd.args(&["--amount", &trust_amount.to_string()[..]]);
     }
@@ -270,6 +274,7 @@ fn sq_link_add_retract() -> Result<()> {
         Command::cargo_bin("sq")
             .unwrap()
             .arg("--no-cert-store")
+            .arg("--no-key-store")
             .arg("sign")
             .args(["--signer-file", &data.key_file])
             .args(["--output", &data.sig_file])
@@ -403,6 +408,7 @@ fn sq_link_add_retract() -> Result<()> {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("sign")
         .args(["--signer-file", &ed_pgp])
         .args(["--output", &ed_sig_file])
@@ -598,7 +604,7 @@ fn sq_link_update_detection() -> Result<()> {
     Ok(())
 }
 
-// Check that sq link add --temporary works.
+// Check that sq pki link add --temporary works.
 #[test]
 fn sq_link_add_temporary() -> Result<()> {
     let dir = TempDir::new()?;
@@ -618,6 +624,7 @@ fn sq_link_add_temporary() -> Result<()> {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("sign")
         .args(["--signer-file", &alice_pgp])
         .args(["--output", &alice_sig_file])
