@@ -20,6 +20,9 @@ use openpgp::policy::StandardPolicy;
 use openpgp::serialize::stream::{Message, Signer, Compressor, LiteralWriter};
 use openpgp::serialize::Serialize;
 
+mod common;
+use common::Sq;
+
 const P: &StandardPolicy = &StandardPolicy::new();
 
 fn artifact(filename: &str) -> String {
@@ -35,6 +38,7 @@ fn sq_sign() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("sign")
         .args(["--signer-file", &artifact("keys/dennis-simon-anton-private.pgp")])
         .args(["--output", &sig.to_string_lossy()])
@@ -70,6 +74,7 @@ fn sq_sign() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("verify")
         .args(["--signer-file", &artifact("keys/dennis-simon-anton.pgp")])
         .arg(&*sig.to_string_lossy())
@@ -86,6 +91,7 @@ fn sq_sign_with_notations() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("sign")
         .args(["--signer-file", &artifact("keys/dennis-simon-anton-private.pgp")])
         .args(["--output", &sig.to_string_lossy()])
@@ -149,6 +155,7 @@ fn sq_sign_with_notations() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .args(["--known-notation", "foo"])
         .arg("verify")
         .args(["--signer-file", &artifact("keys/dennis-simon-anton.pgp")])
@@ -166,6 +173,7 @@ fn sq_sign_append() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("sign")
         .args(["--signer-file", &artifact("keys/dennis-simon-anton-private.pgp")])
         .args(["--output", &sig0.to_string_lossy()])
@@ -201,6 +209,7 @@ fn sq_sign_append() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("verify")
         .args(["--signer-file", &artifact("keys/dennis-simon-anton.pgp")])
         .arg(&*sig0.to_string_lossy())
@@ -212,6 +221,7 @@ fn sq_sign_append() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("sign")
         .arg("--append")
         .args(["--signer-file", &artifact("keys/erika-corinna-daniela-simone-antonia-nistp256-private.pgp")])
@@ -262,6 +272,7 @@ fn sq_sign_append() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("verify")
         .args(["--signer-file", &artifact("keys/dennis-simon-anton.pgp")])
         .arg(&*sig1.to_string_lossy())
@@ -270,6 +281,7 @@ fn sq_sign_append() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("verify")
         .args(["--signer-file", &artifact("keys/erika-corinna-daniela-simone-antonia-nistp256.pgp")])
         .arg(&*sig1.to_string_lossy())
@@ -332,6 +344,7 @@ fn sq_sign_append_on_compress_then_sign() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("verify")
         .args(["--signer-file", &artifact("keys/dennis-simon-anton.pgp")])
         .arg(&*sig0.to_string_lossy())
@@ -343,6 +356,7 @@ fn sq_sign_append_on_compress_then_sign() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("sign")
         .arg("--append")
         .args(["--signer-file", &artifact("keys/erika-corinna-daniela-simone-antonia-nistp256-private.pgp")])
@@ -396,6 +410,7 @@ fn sq_sign_append_on_compress_then_sign() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("verify")
         .args(["--signer-file", &artifact("keys/dennis-simon-anton.pgp")])
         .arg(&*sig0.to_string_lossy())
@@ -405,6 +420,7 @@ fn sq_sign_append_on_compress_then_sign() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("verify")
         .args(["--signer-file", &artifact("keys/erika-corinna-daniela-simone-antonia-nistp256.pgp")])
         .arg(&*sig0.to_string_lossy())
@@ -421,6 +437,7 @@ fn sq_sign_detached() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("sign")
         .arg("--detached")
         .args(["--signer-file", &artifact("keys/dennis-simon-anton-private.pgp")])
@@ -446,6 +463,7 @@ fn sq_sign_detached() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("verify")
         .args(["--signer-file", &artifact("keys/dennis-simon-anton.pgp")])
         .args(["--detached", &sig.to_string_lossy()])
@@ -463,6 +481,7 @@ fn sq_sign_detached_append() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("sign")
         .arg("--detached")
         .args(["--signer-file", &artifact("keys/dennis-simon-anton-private.pgp")])
@@ -488,6 +507,7 @@ fn sq_sign_detached_append() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("verify")
         .args(["--signer-file", &artifact("keys/dennis-simon-anton.pgp")])
         .args(["--detached", &sig.to_string_lossy()])
@@ -499,6 +519,7 @@ fn sq_sign_detached_append() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("sign")
         .arg("--detached")
         .args(["--signer-file", &artifact("keys/erika-corinna-daniela-simone-antonia-nistp256-private.pgp")])
@@ -511,6 +532,7 @@ fn sq_sign_detached_append() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("sign")
         .arg("--detached")
         .arg("--append")
@@ -542,6 +564,7 @@ fn sq_sign_detached_append() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("verify")
         .args(["--signer-file", &artifact("keys/dennis-simon-anton.pgp")])
         .args(["--detached", &sig.to_string_lossy()])
@@ -552,6 +575,7 @@ fn sq_sign_detached_append() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("verify")
         .args(["--signer-file", &artifact("keys/erika-corinna-daniela-simone-antonia-nistp256.pgp")])
         .args(["--detached", &sig.to_string_lossy()])
@@ -564,6 +588,7 @@ fn sq_sign_detached_append() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("sign")
         .arg("--detached")
         .arg("--append")
@@ -602,6 +627,7 @@ fn sq_sign_append_a_notarization() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("sign")
         .arg("--append")
         .args(["--signer-file", &artifact("keys/erika-corinna-daniela-simone-antonia-nistp256-private.pgp")])
@@ -663,6 +689,7 @@ fn sq_sign_append_a_notarization() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("verify")
         .args(["--signer-file", &artifact("keys/neal.pgp")])
         .arg(&*sig0.to_string_lossy())
@@ -671,6 +698,7 @@ fn sq_sign_append_a_notarization() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("verify")
         .args(["--signer-file", &artifact("keys/emmelie-dorothea-dina-samantha-awina-ed25519.pgp")])
         .arg(&*sig0.to_string_lossy())
@@ -679,6 +707,7 @@ fn sq_sign_append_a_notarization() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("verify")
         .args(["--signer-file", &artifact("keys/erika-corinna-daniela-simone-antonia-nistp256.pgp")])
         .arg(&*sig0.to_string_lossy())
@@ -696,6 +725,7 @@ fn sq_sign_notarize() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("sign")
         .arg("--notarize")
         .args(["--signer-file", &artifact("keys/erika-corinna-daniela-simone-antonia-nistp256-private.pgp")])
@@ -745,6 +775,7 @@ fn sq_sign_notarize() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("verify")
         .args(["--signer-file", &artifact("keys/neal.pgp")])
         .arg(&*sig0.to_string_lossy())
@@ -753,6 +784,7 @@ fn sq_sign_notarize() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("verify")
         .args(["--signer-file", &artifact("keys/erika-corinna-daniela-simone-antonia-nistp256.pgp")])
         .arg(&*sig0.to_string_lossy())
@@ -770,6 +802,7 @@ fn sq_sign_notarize_a_notarization() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("sign")
         .arg("--notarize")
         .args(["--signer-file", &artifact("keys/erika-corinna-daniela-simone-antonia-nistp256-private.pgp")])
@@ -831,6 +864,7 @@ fn sq_sign_notarize_a_notarization() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("verify")
         .args(["--signer-file", &artifact("keys/neal.pgp")])
         .arg(&*sig0.to_string_lossy())
@@ -839,6 +873,7 @@ fn sq_sign_notarize_a_notarization() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("verify")
         .args(["--signer-file", &artifact("keys/emmelie-dorothea-dina-samantha-awina-ed25519.pgp")])
         .arg(&*sig0.to_string_lossy())
@@ -847,6 +882,7 @@ fn sq_sign_notarize_a_notarization() {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("verify")
         .args(["--signer-file", &artifact("keys/erika-corinna-daniela-simone-antonia-nistp256.pgp")])
         .arg(&*sig0.to_string_lossy())
@@ -878,7 +914,7 @@ fn sq_multiple_signers() -> Result<()> {
     // Sign message.
     let assertion = Command::cargo_bin("sq")?
         .args([
-            "--no-cert-store",
+            "--no-cert-store", "--no-key-store",
             "sign",
             "--signer-file", alice_pgp.to_str().unwrap(),
             "--signer-file", &bob_pgp.to_str().unwrap(),
@@ -939,7 +975,7 @@ fn sq_sign_using_cert_store() -> Result<()> {
     let mut cmd = Command::cargo_bin("sq")?;
     cmd.args(["--cert-store", &certd,
               "key", "generate",
-              "--expiry", "never",
+              "--expiration", "never",
               "--userid", "<alice@example.org>",
               "--output", &alice_pgp]);
     cmd.assert().success();
@@ -957,6 +993,7 @@ fn sq_sign_using_cert_store() -> Result<()> {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("sign")
         .args(["--signer-file", &alice_pgp])
         .args(["--output", &msg_pgp])
@@ -998,6 +1035,7 @@ fn sq_sign_using_cert_store() -> Result<()> {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("verify")
         .args(["--signer-file", &alice_pgp])
         .arg(&msg_pgp)
@@ -1008,6 +1046,7 @@ fn sq_sign_using_cert_store() -> Result<()> {
     // certificate or use a certificate store.
     let mut cmd = Command::cargo_bin("sq").unwrap();
     cmd.arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("verify")
         .arg(&msg_pgp);
     let output = cmd.output().expect("success");
@@ -1098,8 +1137,9 @@ fn sq_verify_wot() -> Result<()> {
     {
         let mut cmd = Command::cargo_bin("sq").expect("have sq");
         cmd.args(["--no-cert-store",
+                  "--no-key-store",
                   "key", "generate",
-                  "--expiry", "never",
+                  "--expiration", "never",
                   "--output", file]);
         for userid in userids.iter() {
             cmd.args(["--userid", userid]);
@@ -1120,6 +1160,7 @@ fn sq_verify_wot() -> Result<()> {
                      msg_pgp: &str|
     {
         let mut cmd = Command::cargo_bin("sq").expect("have sq");
+        cmd.arg("--no-key-store");
         if let Some(cert_store) = cert_store {
             cmd.args(&["--cert-store", cert_store]);
         } else {
@@ -1149,7 +1190,7 @@ fn sq_verify_wot() -> Result<()> {
     {
         let mut cmd = Command::cargo_bin("sq").expect("have sq");
         cmd.args(&["--cert-store", cert_store]);
-        cmd.args(&["pki", "certify", key, cert, userid]);
+        cmd.args(&["pki", "certify", "--certifier-file", key, cert, userid]);
         if let Some(trust_amount) = trust_amount {
             cmd.args(&["--amount", &trust_amount.to_string()[..]]);
         }
@@ -1186,6 +1227,7 @@ fn sq_verify_wot() -> Result<()> {
     Command::cargo_bin("sq")
         .unwrap()
         .arg("--no-cert-store")
+        .arg("--no-key-store")
         .arg("sign")
         .args(["--signer-file", &bob_pgp])
         .args(["--signer-file", &carol_pgp])
@@ -1304,3 +1346,35 @@ fn sq_verify_wot() -> Result<()> {
 
     Ok(())
 }
+
+#[test]
+fn sq_sign_keyring() {
+    // Check that we can provide the secret key material via
+    // --keyring.
+
+    let sq = Sq::new();
+
+    let (_alice, alice_pgp, _alice_rev)
+        = sq.key_generate(&[], &["alice"]);
+
+    let mut alice_pub = alice_pgp.clone();
+    alice_pub.set_extension("pub");
+
+    sq.toolbox_extract_cert(&alice_pgp, Some(&*alice_pub));
+
+    // We pass the secret key material via --keyring.  This should
+    // work.
+    let mut cmd = sq.command();
+    cmd.arg("--keyring").arg(&alice_pgp)
+        .arg("sign")
+        .arg("--signer-file").arg(&alice_pub);
+
+    sq.run(cmd, Some(true));
+
+    // If we don't pass the secret key material, this should fail.
+    let mut cmd = sq.command();
+    cmd.arg("sign")
+        .arg("--signer-file").arg(&alice_pub);
+
+    sq.run(cmd, Some(false));
+}
